from bancor_research.bancor_emulator.solidity import uint256

'''
* @dev Standard math utilities missing in the Solidity language.
'''
class Math:
    '''
        * @dev Returns the largest of two numbers.
    '''
    def max(a, b):
        a, b = uint256(a), uint256(b)
        return a if a >= b else b;

    '''
        * @dev Returns the smallest of two numbers.
    '''
    def min(a, b):
        a, b = uint256(a), uint256(b)
        return a if a < b else b;

    '''
        * @dev Returns the average of two numbers. The result is rounded towards
        * zero.
    '''
    def average(a, b):
        # (a + b) / 2 can overflow.
        a, b = uint256(a), uint256(b)
        return (a & b) + (a ^ b) / 2;

    '''
        * @dev Returns the ceiling of the division of two numbers.
        *
        * This differs from standard division with `/` in that it rounds up instead
        * of rounding down.
    '''
    def ceilDiv(a, b):
        # (a + b - 1) / b can overflow on addition, so we distribute.
        a, b = uint256(a), uint256(b)
        return a / b + (0 if a % b == 0 else 1);



from bancor_research.bancor_emulator.utils import contract


class Vault(contract):
    '''
     * @dev a "virtual" constructor that is only used to set immutable state variables
    '''

    def __init__(self, initBNTGovernance, initVBNTGovernance) -> None:
        contract.__init__(self)

        self._bntGovernance = initBNTGovernance;
        self._bnt = initBNTGovernance.token();
        self._vbntGovernance = initVBNTGovernance;
        self._vbnt = initVBNTGovernance.token();

    '''
     * @inheritdoc IVault
    '''

    def withdrawFunds(self,
                      token,
                      target,
                      amount
                      ) -> None:
        if (amount == 0):
            return;

        token.transfer(target, amount);

    '''
     * @inheritdoc IVault
    '''

    def burn(self, token, amount) -> None:
        if (amount == 0):
            return;

        if (token is (self._bnt)):
            self._bntGovernance.burn(amount);
        elif (token is (self._vbnt)):
            self._vbntGovernance.burn(amount);
        else:
            token.burn(amount);



from bancor_research.bancor_emulator.solidity import uint, uint256, mapping, address, payable, revert
from bancor_research.bancor_emulator.utils import contract, parse

from bancor_research.bancor_emulator.EnumerableSet import EnumerableSet
from bancor_research.bancor_emulator.Constants import PPM_RESOLUTION
from bancor_research.bancor_emulator.Time import Time
from bancor_research.bancor_emulator.MathEx import MathEx
from bancor_research.bancor_emulator.PoolCollection import PoolCollection as IPoolCollection

'''
 * @dev Bancor Network contract
'''
class BancorNetwork(contract, Time):
    class TradeParams:
        def __init__(self, x = None) -> None:
            self.amount = parse(uint256, x, 'amount');
            self.limit = parse(uint256, x, 'limit');
            self.bySourceAmount = parse(bool, x, 'bySourceAmount');

    class TradeResult:
        def __init__(self, x = None) -> None:
            self.sourceAmount = parse(uint256, x, 'sourceAmount');
            self.targetAmount = parse(uint256, x, 'targetAmount');
            self.tradingFeeAmount = parse(uint256, x, 'tradingFeeAmount');
            self.networkFeeAmount = parse(uint256, x, 'networkFeeAmount');

    class TradeTokens:
        def __init__(self, x = None) -> None:
            self.sourceToken = parse(address, x, 'sourceToken');
            self.targetToken = parse(address, x, 'targetToken');

    class TraderInfo:
        def __init__(self, x = None) -> None:
            self.trader = parse(address, x, 'trader');
            self.beneficiary = parse(address, x, 'beneficiary');

    '''
     * @dev a "virtual" constructor that is only used to set immutable state variables
    '''
    def __init__(self,
        initBNTGovernance,
        initVBNTGovernance,
        initNetworkSettings,
        initMasterVault,
        initExternalProtectionVault,
        initBNTPoolToken
    ) -> None:
        contract.__init__(self)
        Time.__init__(self)

        self._bntGovernance = initBNTGovernance;
        self._bnt = initBNTGovernance.token();
        self._vbntGovernance = initVBNTGovernance;
        self._vbnt = initVBNTGovernance.token();

        self._networkSettings = initNetworkSettings;
        self._masterVault = initMasterVault;
        self._externalProtectionVault = initExternalProtectionVault;
        self._bntPoolToken = initBNTPoolToken;

    '''
     * @dev fully initializes the contract and its parents
    '''
    def initialize(self,
         initBNTPool,
         initPendingWithdrawals,
         initPoolMigrator
    ) -> None:
        self.__BancorNetwork_init(initBNTPool, initPendingWithdrawals, initPoolMigrator);

    # solhint-disable func-name-mixedcase

    '''
     * @dev initializes the contract and its parents
    '''
    def __BancorNetwork_init(self,
         initBNTPool,
         initPendingWithdrawals,
         initPoolMigrator
    ) -> None:
        self.__BancorNetwork_init_unchained(initBNTPool, initPendingWithdrawals, initPoolMigrator);

    '''
     * @dev performs contract-specific initialization
    '''
    def __BancorNetwork_init_unchained(self,
         initBNTPool,
         initPendingWithdrawals,
         initPoolMigrator
    ) -> None:
        self._bntPool = initBNTPool;
        self._pendingWithdrawals = initPendingWithdrawals;
        self._poolMigrator = initPoolMigrator;

        # the set of all valid pool collections
        self._poolCollections = EnumerableSet();

        # the set of all pools
        self._liquidityPools = EnumerableSet();

        # a mapping between pools and their respective pool collections
        self._collectionByPool = mapping(lambda: address(0));

        # the pending network fee amount to be burned by the vortex
        self._pendingNetworkFeeAmount = uint256();

    # solhint-enable func-name-mixedcase

    '''
     * @inheritdoc Upgradeable
    '''
    def version(self) -> (int):
        return 7;

    '''
     * @dev returns the pending network fee amount to be burned by the vortex
    '''
    def pendingNetworkFeeAmount(self) -> (uint):
        return self._pendingNetworkFeeAmount.clone();

    '''
     * @dev registers new pool collection with the network
     *
     * requirements:
     *
     * - the caller must be the admin of the contract
    '''
    def registerPoolCollection(self, newPoolCollection) -> None:
        # verify that there is no pool collection of the same type and version
        newPoolType = newPoolCollection.poolType();
        newPoolVersion = newPoolCollection.version();

        poolCollection = self._findPoolCollection(newPoolType, newPoolVersion);
        if (poolCollection != address(0) or not self._poolCollections.add(address(newPoolCollection))):
            revert("AlreadyExists");

    '''
     * @dev unregisters an existing pool collection from the network
     *
     * requirements:
     *
     * - the caller must be the admin of the contract
    '''
    def unregisterPoolCollection(self, poolCollection) -> None:
        # verify that no pools are associated with the specified pool collection
        if (poolCollection.poolCount() != 0):
            revert("NotEmpty");

        if (not self._poolCollections.remove(address(poolCollection))):
            revert("DoesNotExist");

        self._setAccessRoles(poolCollection, False);

    '''
     * @inheritdoc IBancorNetwork
    '''
    def poolCollections(self) -> (list):
        return self._poolCollections.values();

    '''
     * @inheritdoc IBancorNetwork
    '''
    def liquidityPools(self) -> (list):
        return self._liquidityPools.values();

    '''
     * @inheritdoc IBancorNetwork
    '''
    def collectionByPool(self, pool) -> (IPoolCollection):
        return self._collectionByPool[pool];

    '''
     * @inheritdoc IBancorNetwork
    '''
    def createPools(self, tokens, poolCollection) -> None:
        if (not self._poolCollections.contains(address(poolCollection))):
            revert("DoesNotExist");

        length = len(tokens);
        for i in range(length):
            self._createPool(tokens[i], poolCollection);

    '''
     * @dev creates a new pool
    '''
    def _createPool(self, token, poolCollection) -> None:
        if (token is (self._bnt)):
            revert("InvalidToken");

        if (not self._liquidityPools.add(address(token))):
            revert("AlreadyExists");

        # this is where the magic happens...
        poolCollection.createPool(token);

        # add the pool collection to the reverse pool collection lookup
        self._collectionByPool[token] = poolCollection;

    '''
     * @inheritdoc IBancorNetwork
    '''
    def migratePools(self, pools, newPoolCollection) -> None:
        if (not self._poolCollections.contains(address(newPoolCollection))):
            revert("DoesNotExist");

        length = len(pools);
        for i in range(length):
            pool = pools[i];

            # request the pool migrator to migrate the pool to the new pool collection
            self._poolMigrator.migratePool(pool, newPoolCollection);

            prevPoolCollection = self._collectionByPool[pool];

            # update the mapping between pools and their respective pool collections
            self._collectionByPool[pool] = newPoolCollection;

    '''
     * @inheritdoc IBancorNetwork
    '''
    def depositFor(self,
        provider,
        pool,
        tokenAmount
    ) -> (uint):
        return self._depositFor(provider, pool, tokenAmount, self.msg_sender);

    '''
     * @inheritdoc IBancorNetwork
    '''
    def deposit(self, pool, tokenAmount) -> (uint):
        return self._depositFor(self.msg_sender, pool, tokenAmount, self.msg_sender);

    '''
     * @inheritdoc IBancorNetwork
    '''
    def initWithdrawal(self, poolToken, poolTokenAmount) -> (uint):
        return self._initWithdrawal(self.msg_sender, poolToken, poolTokenAmount);

    '''
     * @inheritdoc IBancorNetwork
    '''
    def cancelWithdrawal(self, id) -> (uint):
        return self._pendingWithdrawals.cancelWithdrawal(self.msg_sender, id);

    '''
     * @inheritdoc IBancorNetwork
    '''
    def withdraw(self, id) -> (uint):
        provider = self.msg_sender;
        contextId = self._withdrawContextId(id, provider);

        # complete the withdrawal and claim the locked pool tokens
        completedRequest = self._pendingWithdrawals.completeWithdrawal(contextId, provider, id);

        if (completedRequest.poolToken == self._bntPoolToken):
            return self._withdrawBNT(contextId, provider, completedRequest);

        return self._withdrawBaseToken(contextId, provider, completedRequest);

    '''
     * @inheritdoc IBancorNetwork
    '''
    def tradeBySourceAmount(self,
        sourceToken,
        targetToken,
        sourceAmount,
        minReturnAmount,
        deadline,
        beneficiary
    ) -> (uint):
        self._verifyTradeParams(sourceToken, targetToken, sourceAmount, minReturnAmount, deadline);

        return \
            self._trade(
                self.TradeTokens({ 'sourceToken': sourceToken, 'targetToken': targetToken }),
                self.TradeParams({ 'bySourceAmount': True, 'amount': sourceAmount, 'limit': minReturnAmount }),
                self.TraderInfo({ 'trader': self.msg_sender, 'beneficiary': beneficiary }),
                deadline
            );

    '''
     * @inheritdoc IBancorNetwork
    '''
    def tradeByTargetAmount(self,
        sourceToken,
        targetToken,
        targetAmount,
        maxSourceAmount,
        deadline,
        beneficiary
    ) -> (uint):
        self._verifyTradeParams(sourceToken, targetToken, targetAmount, maxSourceAmount, deadline);

        return \
            self._trade(
                self.TradeTokens({ 'sourceToken': sourceToken, 'targetToken': targetToken }),
                self.TradeParams({ 'bySourceAmount': False, 'amount': targetAmount, 'limit': maxSourceAmount }),
                self.TraderInfo({ 'trader': self.msg_sender, 'beneficiary': beneficiary }),
                deadline
            );

    '''
     * @inheritdoc IBancorNetwork
    '''
    def flashLoan(self,
        token,
        amount,
        recipient,
        data
    ) -> None:
        if (not token is (self._bnt) and not self._networkSettings.isTokenWhitelisted(token)):
            revert("NotWhitelisted");

        feeAmount = MathEx.mulDivF(amount, self._networkSettings.flashLoanFeePPM(token), PPM_RESOLUTION);

        # save the current balance
        prevBalance = token.balanceOf(address(self));

        # transfer the amount from the master vault to the recipient
        self._masterVault.withdrawFunds(token, payable(address(recipient)), amount);

        # invoke the recipient's callback
        recipient.onFlashLoan(self.msg_sender, token.toIERC20(), amount, feeAmount, data);

        # ensure that the tokens + fee have been deposited back to the network
        returnedAmount = token.balanceOf(address(self)) - prevBalance;
        if (returnedAmount < amount + feeAmount):
            revert("InsufficientFlashLoanReturn");

        # transfer the amount and the fee back to the vault
        token.transfer(payable(address(self._masterVault)), returnedAmount);

        # notify the pool of accrued fees
        if (token is (self._bnt)):
            cachedBNTPool = self._bntPool;

            cachedBNTPool.onFeesCollected(token, feeAmount, False);
        else:
            # get the pool and verify that it exists
            poolCollection = self._poolCollection(token);
            poolCollection.onFeesCollected(token, feeAmount);

    '''
     * @inheritdoc IBancorNetwork
    '''
    def migrateLiquidity(self,
        token,
        provider,
        amount,
        availableAmount,
        originalAmount
    ) -> None:
        contextId = uint256(0);

        if (token is (self._bnt)):
            self._depositBNTFor(contextId, provider, amount, True, originalAmount);
        else:
            self._depositBaseTokenFor(contextId, provider, token, amount, availableAmount);

    '''
     * @inheritdoc IBancorNetwork
    '''
    def withdrawNetworkFees(self, recipient) -> (uint):
        currentPendingNetworkFeeAmount = self._pendingNetworkFeeAmount;
        if (currentPendingNetworkFeeAmount == 0):
            return uint256(0);

        self._pendingNetworkFeeAmount = uint256(0);

        self._masterVault.withdrawFunds(address(self._bnt), payable(recipient), currentPendingNetworkFeeAmount);

        return currentPendingNetworkFeeAmount;

    '''
     * @dev generates context ID for a deposit request
    '''
    def _depositContextId(self,
        provider,
        pool,
        tokenAmount,
        caller
    ) -> (uint):
        return uint256(0);

    '''
     * @dev generates context ID for a withdraw request
    '''
    def _withdrawContextId(self, id, caller) -> (uint):
        return uint256(0);

    '''
     * @dev deposits liquidity for the specified provider from caller
     *
     * requirements:
     *
     * - the caller must have approved the network to transfer the liquidity tokens on its behalf
    '''
    def _depositFor(self,
        provider,
        pool,
        tokenAmount,
        caller
    ) -> (uint):
        contextId = self._depositContextId(provider, pool, tokenAmount, caller);

        if (pool is (self._bnt)):
            return self._depositBNTFor(contextId, provider, tokenAmount, caller, False, 0);

        return self._depositBaseTokenFor(contextId, provider, pool, tokenAmount, caller, tokenAmount);

    '''
     * @dev deposits BNT liquidity for the specified provider from caller
     *
     * requirements:
     *
     * - the caller must have approved the network to transfer BNT on its behalf
    '''
    def _depositBNTFor(self,
        contextId,
        provider,
        bntAmount,
        caller,
        isMigrating,
        originalAmount
    ) -> (uint):
        cachedBNTPool = self._bntPool;

        # transfer the tokens from the caller to the BNT pool
        self._bnt.transferFrom(caller, address(cachedBNTPool), bntAmount);

        # process BNT pool deposit
        return cachedBNTPool.depositFor(contextId, provider, bntAmount, isMigrating, originalAmount);

    '''
     * @dev deposits base token liquidity for the specified provider from sender
     *
     * requirements:
     *
     * - the caller must have approved the network to transfer base tokens to on its behalf
    '''
    def _depositBaseTokenFor(self,
         contextId,
         provider,
         pool,
         tokenAmount,
         caller,
         availableAmount
    ) -> (uint):
        # transfer the tokens from the sender to the vault
        self._depositToMasterVault(pool, caller, availableAmount);

        # get the pool collection that managed this pool
        poolCollection = self._poolCollection(pool);

        # process deposit to the base token pool (includes the native token pool)
        return poolCollection.depositFor(contextId, provider, pool, tokenAmount);

    '''
     * @dev handles BNT withdrawal
    '''
    def _withdrawBNT(self,
        contextId,
        provider,
        completedRequest
    ) -> (uint):
        cachedBNTPool = self._bntPool;

        # transfer the pool tokens to from the pending withdrawals contract to the BNT pool
        completedRequest.poolToken.transferFrom(
            address(self._pendingWithdrawals),
            address(cachedBNTPool),
            completedRequest.poolTokenAmount
        );

        # transfer vBNT from the caller to the BNT pool
        self._vbnt.transferFrom(provider, address(cachedBNTPool), completedRequest.poolTokenAmount);

        # call withdraw on the BNT pool
        return \
            cachedBNTPool.withdraw(
                contextId,
                provider,
                completedRequest.poolTokenAmount,
                completedRequest.reserveTokenAmount
            );

    '''
     * @dev handles base token withdrawal
    '''
    def _withdrawBaseToken(self,
        contextId,
        provider,
        completedRequest
    ) -> (uint):
        pool = completedRequest.poolToken.reserveToken();

        # get the pool collection that manages this pool
        poolCollection = self._poolCollection(pool);

        # transfer the pool tokens to from the pending withdrawals contract to the pool collection
        completedRequest.poolToken.transferFrom(
            address(self._pendingWithdrawals),
            address(poolCollection),
            completedRequest.poolTokenAmount
        );

        # call withdraw on the base token pool - returns the amounts/breakdown
        return \
            poolCollection.withdraw(
                contextId,
                provider,
                pool,
                completedRequest.poolTokenAmount,
                completedRequest.reserveTokenAmount
            );

    '''
     * @dev verifies that the provided trade params are valid
    '''
    def _verifyTradeParams(self,
        sourceToken,
        targetToken,
        amount,
        limit,
        deadline
    ) -> None:
        if (sourceToken == targetToken):
            revert("InvalidToken");

        if (deadline < self._time()):
            revert("DeadlineExpired");

    '''
     * @dev performs a trade by providing either the source or target 'amount':
     *
     * - when trading by the source amount, the amount represents the source amount and the limit is the minimum return
     *   amount
     * - when trading by the target amount, the amount represents the target amount and the limit is the maximum source
     *   amount
     *
     * requirements:
     *
     * - the caller must have approved the network to transfer the source tokens on its behalf (except for in the
     *   native token case)
    '''
    def _trade(self,
        tokens,
        params,
        traderInfo,
        deadline
    ) -> (uint):
        # ensure the beneficiary is set
        if (traderInfo.beneficiary == address(0)):
            traderInfo.beneficiary = traderInfo.trader;

        contextId = uint256(0);

        # perform either a single or double hop trade, based on the source and the target pool
        firstHopTradeResult = self.TradeResult();
        lastHopTradeResult = self.TradeResult();
        networkFeeAmount = uint256();

        if (tokens.sourceToken is (self._bnt)):
            lastHopTradeResult = self._tradeBNT(contextId, tokens.targetToken, True, params);

            firstHopTradeResult = lastHopTradeResult;

            networkFeeAmount = lastHopTradeResult.networkFeeAmount;
        elif (tokens.targetToken is (self._bnt)):
            lastHopTradeResult = self._tradeBNT(contextId, tokens.sourceToken, False, params);

            firstHopTradeResult = lastHopTradeResult;

            networkFeeAmount = lastHopTradeResult.networkFeeAmount;
        else:
            (firstHopTradeResult, lastHopTradeResult) = self._tradeBaseTokens(contextId, tokens, params);

            networkFeeAmount = firstHopTradeResult.networkFeeAmount + lastHopTradeResult.networkFeeAmount;

        # transfer the tokens from the trader to the vault
        self._depositToMasterVault(tokens.sourceToken, traderInfo.trader, firstHopTradeResult.sourceAmount);

        # transfer the target tokens/native token to the beneficiary
        self._masterVault.withdrawFunds(
            tokens.targetToken,
            payable(traderInfo.beneficiary),
            lastHopTradeResult.targetAmount
        );

        # update the pending network fee amount to be burned by the vortex
        self._pendingNetworkFeeAmount += networkFeeAmount;

        return lastHopTradeResult.targetAmount if params.bySourceAmount else firstHopTradeResult.sourceAmount;

    '''
     * @dev performs a single hop between BNT and a base token trade by providing either the source or the target amount
     *
     * - when trading by the source amount, the amount represents the source amount and the limit is the minimum return
     *   amount
     * - when trading by the target amount, the amount represents the target amount and the limit is the maximum source
     *   amount
    '''
    def _tradeBNT(self,
        contextId,
        pool,
        fromBNT,
        params
    ) -> (TradeResult):
        tokens = \
            self.TradeTokens({ 'sourceToken': address(self._bnt), 'targetToken': pool }) if fromBNT else \
            self.TradeTokens({ 'sourceToken': pool, 'targetToken': address(self._bnt) });

        tradeAmountsAndFee = \
            self._poolCollection(pool).tradeBySourceAmount(
                contextId,
                tokens.sourceToken,
                tokens.targetToken,
                params.amount,
                params.limit
            ) if params.bySourceAmount else \
            self._poolCollection(pool).tradeByTargetAmount(
                contextId,
                tokens.sourceToken,
                tokens.targetToken,
                params.amount,
                params.limit
            );

        # if the target token is BNT, notify the BNT pool on collected fees (which shouldn't include the network fee
        # amount, so we have to deduct it explicitly from the full trading fee amount)
        if (not fromBNT):
            self._bntPool.onFeesCollected(
                pool,
                tradeAmountsAndFee.tradingFeeAmount - tradeAmountsAndFee.networkFeeAmount,
                True
            );

        return \
            self.TradeResult({
                'sourceAmount': params.amount if params.bySourceAmount else tradeAmountsAndFee.amount,
                'targetAmount': tradeAmountsAndFee.amount if params.bySourceAmount else params.amount,
                'tradingFeeAmount': tradeAmountsAndFee.tradingFeeAmount,
                'networkFeeAmount': tradeAmountsAndFee.networkFeeAmount
            });

    '''
     * @dev performs a double hop trade between two base tokens by providing either the source or the target amount
     *
     * - when trading by the source amount, the amount represents the source amount and the limit is the minimum return
     *   amount
     * - when trading by the target amount, the amount represents the target amount and the limit is the maximum source
     *   amount
    '''
    def _tradeBaseTokens(self,
        contextId,
        tokens,
        params
    ):# -> (TradeResult, TradeResult):
        if (params.bySourceAmount):
            sourceAmount = params.amount;
            minReturnAmount = params.limit;

            # trade source tokens to BNT (while accepting any return amount)
            targetHop1 = self._tradeBNT(
                contextId,
                tokens.sourceToken,
                False,
                self.TradeParams({ 'bySourceAmount': True, 'amount': sourceAmount, 'limit': 1 })
            );

            # trade the received BNT target amount to target tokens (while respecting the minimum return amount)
            targetHop2 = self._tradeBNT(
                contextId,
                tokens.targetToken,
                True,
                self.TradeParams({ 'bySourceAmount': True, 'amount': targetHop1.targetAmount, 'limit': minReturnAmount })
            );

            return (targetHop1, targetHop2);

        targetAmount = params.amount;
        maxSourceAmount = params.limit;

        # trade any amount of BNT to get the requested target amount (we will use the actual traded amount to restrict
        # the trade from the source)
        sourceHop2 = self._tradeBNT(
            contextId,
            tokens.targetToken,
            True,
            self.TradeParams({ 'bySourceAmount': False, 'amount': targetAmount, 'limit': uint256.max })
        );

        # trade source tokens to the required amount of BNT (while respecting the maximum source amount)
        sourceHop1 = self._tradeBNT(
            contextId,
            tokens.sourceToken,
            False,
            self.TradeParams({ 'bySourceAmount': False, 'amount': sourceHop2.sourceAmount, 'limit': maxSourceAmount })
        );

        return (sourceHop1, sourceHop2);

    '''
     * @dev deposits tokens to the master vault and verifies that msg.value corresponds to its type
    '''
    def _depositToMasterVault(self,
        token,
        caller,
        amount
    ) -> None:
        token.transferFrom(caller, address(self._masterVault), amount);

    '''
     * @dev verifies that the specified pool is managed by a valid pool collection and returns it
    '''
    def _poolCollection(self, token) -> (IPoolCollection):
        # verify that the pool is managed by a valid pool collection
        poolCollection = self._collectionByPool[token];
        if (address(poolCollection) == address(0)):
            revert("InvalidToken");

        return poolCollection;

    '''
     * @dev initiates liquidity withdrawal
    '''
    def _initWithdrawal(self,
        provider,
        poolToken,
        poolTokenAmount
    ) -> (uint):
        if (poolToken != self._bntPoolToken):
            reserveToken = poolToken.reserveToken();
            if (self._poolCollection(reserveToken).poolToken(reserveToken) != poolToken):
                revert("InvalidPool");

        # transfer the pool tokens from the provider (we aren't using safeTransferFrom, since the PoolToken is a fully
        # compliant ERC20 token contract)
        poolToken.transferFrom(provider, address(self._pendingWithdrawals), poolTokenAmount);

        return self._pendingWithdrawals.initWithdrawal(provider, poolToken, poolTokenAmount);

    '''
     * @dev finds a pool collection with the given type and version
    '''
    def _findPoolCollection(self, poolType, poolVersion) -> (IPoolCollection):
        # note that there's no risk of using an unbounded loop here since the list of all the active pool collections
        # is always going to remain sufficiently small
        length = self._poolCollections.length();
        for i in range(length):
            poolCollection = self._poolCollections.at(i);
            if ((poolCollection.poolType() == poolType and poolCollection.version() == poolVersion)):
                return poolCollection;

        return address(0);



# coding=utf-8
# --------------------------------------------------------------------------------------------------------------------
# Licensed under the MIT LICENSE. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------------------------------
"""Spec, simulation, & emulation modules for Bancor v3."""

import warnings, decimal, pandas
from pydantic.fields import TypeVar

with warnings.catch_warnings():
    warnings.simplefilter("ignore", FutureWarning)
    warnings.simplefilter("ignore", DeprecationWarning)

warnings.filterwarnings("ignore", category=FutureWarning)
warnings.filterwarnings("ignore", category=DeprecationWarning)
warnings.filterwarnings("ignore", category=RuntimeWarning)

decimal.getcontext().prec = len(str(2 ** 512 - 1))
decimal.getcontext().rounding = decimal.ROUND_HALF_EVEN

Decimal = decimal.Decimal

DataFrame = pandas.DataFrame

PandasDataFrame = TypeVar("pandas.core.frame.DataFrame")

SECONDS_PER_DAY = 24 * 60 * 60

class DEFAULT:
    TIMESTAMP = 0

    DECIMALS = 18

    TRADING_FEE    = "1%"
    NETWORK_FEE    = "20%"
    WITHDRAWAL_FEE = "0.25%"

    BNT_FUNDING_LIMIT = "1000000"
    BNT_MIN_LIQUIDITY = "10000"
    EP_VAULT_BALANCE  = "0"

    COOLDOWN_TIME  = 7 * SECONDS_PER_DAY
    NUM_TIMESTAMPS = 30 * SECONDS_PER_DAY

    PRICE_FEEDS_PATH = "https://bancorml.s3.us-east-2.amazonaws.com/price_feeds.parquet"

    PRICE_FEEDS = DataFrame({
        "INDX": (    0.00 for _ in range(NUM_TIMESTAMPS)),
        "vbnt": (    1.00 for _ in range(NUM_TIMESTAMPS)),
        "tkn" : (    2.50 for _ in range(NUM_TIMESTAMPS)),
        "bnt" : (    2.50 for _ in range(NUM_TIMESTAMPS)),
        "link": (   15.00 for _ in range(NUM_TIMESTAMPS)),
        "eth" : ( 2500.00 for _ in range(NUM_TIMESTAMPS)),
        "wbtc": (40000.00 for _ in range(NUM_TIMESTAMPS)),
    })

    USERS = [
        "Alice",
        "Bob",
        "Charlie",
        "Trader",
    ]

    TOKENS = [
        "eth",
        "link",
        "tkn",
        "wbtc",
    ]

DEFAULT.WHITELIST = {
    TOKEN : {
        "decimals": DEFAULT.DECIMALS,
        "trading_fee": DEFAULT.TRADING_FEE,
        "bnt_funding_limit": DEFAULT.BNT_FUNDING_LIMIT,
        "ep_vault_balance": DEFAULT.EP_VAULT_BALANCE,
    }
    for TOKEN in DEFAULT.TOKENS
}

def read_price_feeds(price_feeds_path: str):
    price_feeds = pandas.read_parquet(price_feeds_path)
    price_feeds.columns = [col.lower() for col in price_feeds.columns]
    return price_feeds


__version__ = "3.1.1"



from bancor_research.bancor_emulator.solidity import uint, uint32, uint256, mapping, address, revert
from bancor_research.bancor_emulator.utils import contract, parse

from bancor_research.bancor_emulator.EnumerableSet import EnumerableSet

'''
 * @dev Network Settings contract
'''
class NetworkSettings(contract):
    DEFAULT_FLASH_LOAN_FEE_PPM = uint32(0); # 0%

    class FlashLoanFee:
        def __init__(self, x = None) -> None:
            self.initialized = parse(bool, x, 'initialized');
            self.feePPM = parse(uint32, x, 'feePPM');

    def __init__(self, initBnt) -> None:
        contract.__init__(self)

        self._bnt = initBnt;

    '''
     * @dev fully initializes the contract and its parents
    '''
    def initialize(self) -> None:
        self.__NetworkSettings_init();

    # solhint-disable func-name-mixedcase

    '''
     * @dev initializes the contract and its parents
    '''
    def __NetworkSettings_init(self) -> None:
        self.__NetworkSettings_init_unchained();

    '''
     * @dev performs contract-specific initialization
    '''
    def __NetworkSettings_init_unchained(self) -> None:
        # a set of tokens which are eligible for protection
        self._protectedTokenWhitelist = EnumerableSet();

        # a mapping of BNT funding limits per pool
        self._poolFundingLimits = mapping(lambda: uint256());

        # below that amount, trading is disabled and co-investments use the initial rate
        self._minLiquidityForTrading = uint256();

        # the withdrawal fee (in units of PPM)
        self._withdrawalFeePPM = uint32();

        # the default flash-loan fee (in units of PPM)
        self._defaultFlashLoanFeePPM = uint32();

        # a mapping between pools and their flash-loan fees
        self._flashLoanFees = mapping(lambda: self.FlashLoanFee());

        self._setDefaultFlashLoanFeePPM(self.DEFAULT_FLASH_LOAN_FEE_PPM);

    # solhint-enable func-name-mixedcase

    '''
     * @inheritdoc Upgradeable
    '''
    def version(self) -> (int):
        return 3;

    '''
     * @inheritdoc INetworkSettings
    '''
    def protectedTokenWhitelist(self) -> (list):
        return self._protectedTokenWhitelist.values();

    '''
     * @dev adds a token to the protected tokens whitelist
     *
     * requirements:
     *
     * - the caller must be the admin of the contract
    '''
    def addTokenToWhitelist(self, token) -> None:
        self._addTokenToWhitelist(token);

    '''
     * @dev adds tokens to the protected tokens whitelist
     *
     * requirements:
     *
     * - the caller must be the admin of the contract
    '''
    def addTokensToWhitelist(self, tokens) -> None:
        length = len(tokens.length);

        for i in range(length):
            self._addTokenToWhitelist(tokens[i]);

    '''
     * @dev adds a token to the protected tokens whitelist
    '''
    def _addTokenToWhitelist(self, token) -> None:
        if (not self._protectedTokenWhitelist.add(address(token))):
            revert("AlreadyExists");

    '''
     * @dev removes a token from the protected tokens whitelist
     *
     * requirements:
     *
     * - the caller must be the admin of the contract
    '''
    def removeTokenFromWhitelist(self, token) -> None:
        if (not self._protectedTokenWhitelist.remove(address(token))):
            revert("DoesNotExist");

    '''
     * @inheritdoc INetworkSettings
    '''
    def isTokenWhitelisted(self, token) -> (bool):
        return self._isTokenWhitelisted(token);

    '''
     * @inheritdoc INetworkSettings
    '''
    def poolFundingLimit(self, pool) -> (uint):
        return self._poolFundingLimits[pool].clone();

    '''
     * @dev updates the amount of BNT that the protocol can provide as funding for a specific pool
     *
     * requirements:
     *
     * - the caller must be the admin of the contract
     * - the token must have been whitelisted
    '''
    def setFundingLimit(self, pool, amount) -> None:
        self._setFundingLimit(pool, amount);

    '''
     * @dev updates the amounts of BNT that the protocol can provide as funding for specific pools
     *
     * requirements:
     *
     * - the caller must be the admin of the contract
     * - each one of the tokens must have been whitelisted
    '''
    def setFundingLimits(self, pools, amounts) -> None:
        length = len(pools);
        if (length != len(amounts)):
            revert("InvalidParam");

        for i in range(length):
            self._setFundingLimit(pools[i], amounts[i]);

    '''
     * @dev updates the amount of BNT that the protocol can provide as funding for a specific pool
    '''
    def _setFundingLimit(self, pool, amount) -> None:
        if (not self._isTokenWhitelisted(pool)):
            revert("NotWhitelisted");

        prevPoolFundingLimit = self._poolFundingLimits[pool];
        if (prevPoolFundingLimit == amount):
            return;

        self._poolFundingLimits[pool] = uint256(amount);

    '''
     * @dev adds a token to the protected tokens whitelist,
     * and sets the amount of BNT that the protocol can provide as funding for this pool
     *
     * requirements:
     *
     * - the caller must be the admin of the contract
    '''
    def addTokenToWhitelistWithLimit(self, token, amount) -> None:
        self._addTokenToWhitelist(token);
        self._setFundingLimit(token, amount);

    '''
     * @inheritdoc INetworkSettings
    '''
    def minLiquidityForTrading(self) -> (uint):
        return self._minLiquidityForTrading.clone();

    '''
     * @dev updates the minimum liquidity for trading amount
     *
     * requirements:
     *
     * - the caller must be the admin of the contract
    '''
    def setMinLiquidityForTrading(self, amount) -> None:
        prevMinLiquidityForTrading = self._minLiquidityForTrading;
        if (self._minLiquidityForTrading == amount):
            return;

        self._minLiquidityForTrading = uint256(amount);

    '''
     * @inheritdoc INetworkSettings
    '''
    def withdrawalFeePPM(self) -> (uint):
        return self._withdrawalFeePPM.clone();

    '''
     * @dev sets the withdrawal fee (in units of PPM)
     *
     * requirements:
     *
     * - the caller must be the admin of the contract
    '''
    def setWithdrawalFeePPM(self, newWithdrawalFeePPM) -> None:
        prevWithdrawalFeePPM = self._withdrawalFeePPM;
        if (prevWithdrawalFeePPM == newWithdrawalFeePPM):
            return;

        self._withdrawalFeePPM = uint32(newWithdrawalFeePPM);

    '''
     * @inheritdoc INetworkSettings
    '''
    def defaultFlashLoanFeePPM(self) -> (uint):
        return self._defaultFlashLoanFeePPM.clone();

    '''
     * @dev sets the default flash-loan fee (in units of PPM)
     *
     * requirements:
     *
     * - the caller must be the admin of the contract
    '''
    def setDefaultFlashLoanFeePPM(self, newDefaultFlashLoanFeePPM) -> None:
        self._setDefaultFlashLoanFeePPM(newDefaultFlashLoanFeePPM);

    '''
     * @inheritdoc INetworkSettings
    '''
    def flashLoanFeePPM(self, pool) -> (uint):
        flashLoanFee = self._flashLoanFees[pool];

        return (flashLoanFee.feePPM if flashLoanFee.initialized else self._defaultFlashLoanFeePPM).clone();

    '''
     * @dev sets the flash-loan fee of a given pool
     *
     * requirements:
     *
     * - the caller must be the admin of the contract
     * - the token must have been whitelisted
    '''
    def setFlashLoanFeePPM(self, pool, newFlashLoanFeePPM) -> None:
        if (not pool.isEqual(self._bnt) and not self._isTokenWhitelisted(pool)):
            revert("NotWhitelisted");

        prevFlashLoanFeePPM = self._flashLoanFees[pool].feePPM;
        if (prevFlashLoanFeePPM == newFlashLoanFeePPM):
            return;

        self._flashLoanFees[pool] = self.FlashLoanFee({ 'initialized': True, 'feePPM': newFlashLoanFeePPM });

    '''
     * @dev checks whether a given token is whitelisted
    '''
    def _isTokenWhitelisted(self, token) -> (bool):
        return self._protectedTokenWhitelist.contains(address(token));

    '''
     * @dev sets the default flash-loan fee (in units of PPM)
    '''
    def _setDefaultFlashLoanFeePPM(self, newDefaultFlashLoanFeePPM) -> None:
        prevDefaultFlashLoanFeePPM = self._defaultFlashLoanFeePPM;
        if (prevDefaultFlashLoanFeePPM == newDefaultFlashLoanFeePPM):
            return;

        self._defaultFlashLoanFeePPM = uint32(newDefaultFlashLoanFeePPM);



from bancor_research.bancor_emulator.solidity import uint, uint32, uint256, address, mapping, revert
from bancor_research.bancor_emulator.utils import contract, parse

from bancor_research.bancor_emulator.EnumerableSet import EnumerableSet
from bancor_research.bancor_emulator.Math import Math
from bancor_research.bancor_emulator.Time import Time

class Rewards:
    def __init__(self, x = None) -> None:
        self.lastUpdateTime = parse(uint32, x, 'lastUpdateTime');
        self.rewardPerToken = parse(uint256, x, 'rewardPerToken');

class ProgramData:
    def __init__(self, x = None) -> None:
        self.id = parse(uint256, x, 'id');
        self.pool = parse(address, x, 'pool');
        self.poolToken = parse(address, x, 'poolToken');
        self.rewardsToken = parse(address, x, 'rewardsToken');
        self.isPaused = parse(bool, x, 'isPaused');
        self.startTime = parse(uint32, x, 'startTime');
        self.endTime = parse(uint32, x, 'endTime');
        self.rewardRate = parse(uint256, x, 'rewardRate');
        self.remainingRewards = parse(uint256, x, 'remainingRewards');

class ProviderRewards:
    def __init__(self, x = None) -> None:
        self.rewardPerTokenPaid = parse(uint256, x, 'rewardPerTokenPaid');
        self.pendingRewards = parse(uint256, x, 'pendingRewards');
        self.stakedAmount = parse(uint256, x, 'stakedAmount');

class StakeAmounts:
    def __init__(self, x = None) -> None:
        self.stakedRewardAmount = parse(uint256, x, 'stakedRewardAmount');
        self.poolTokenAmount = parse(uint256, x, 'poolTokenAmount');

'''
 * @dev Standard Rewards contract
'''
class StandardRewards(contract, Time):
    class ClaimData:
        def __init__(self, x = None) -> None:
            self.reward = parse(uint256, x, 'reward');
            self.stakedAmount = parse(uint256, x, 'stakedAmount');

    # since we will be dividing by the total amount of protected tokens in units of wei, we can encounter cases
    # where the total amount in the denominator is higher than the product of the rewards rate and staking duration. In
    # order to avoid this imprecision, we will amplify the reward rate by the units amount.
    REWARD_RATE_FACTOR = uint256(10 ** 18);

    INITIAL_PROGRAM_ID = uint256(1);

    '''
     * @dev a "virtual" constructor that is only used to set immutable state variables
    '''
    def __init__(self,
        initNetwork,
        initNetworkSettings,
        initBNTGovernance,
        initVBNT,
        initBNTPool
    ) -> None:
        contract.__init__(self)
        Time.__init__(self)

        self._network = initNetwork;
        self._networkSettings = initNetworkSettings;
        self._bntGovernance = initBNTGovernance;
        self._bnt = initBNTGovernance.token();
        self._vbnt = initVBNT;
        self._bntPoolToken = initBNTPool.poolToken();

    '''
     * @dev fully initializes the contract and its parents
    '''
    def initialize(self) -> None:
        self.__StandardRewards_init();

    # solhint-disable func-name-mixedcase

    '''
     * @dev initializes the contract and its parents
    '''
    def __StandardRewards_init(self) -> None:
       self.__StandardRewards_init_unchained();

    '''
     * @dev performs contract-specific initialization
    '''
    def __StandardRewards_init_unchained(self) -> None:
        self._nextProgramId = self.INITIAL_PROGRAM_ID.clone();

        # a mapping between providers and the program IDs of the program they are participating in
        self._programIdsByProvider = mapping(lambda: EnumerableSet());

        # a mapping between program IDs and program data
        self._programs = mapping(lambda: ProgramData());

        # a mapping between pools and their latest programs
        self._latestProgramIdByPool = mapping(lambda: uint256());

        # a mapping between programs and their respective rewards data
        self._programRewards = mapping(lambda: Rewards());

        # a mapping between providers, programs and their respective rewards data
        self._providerRewards = mapping(lambda: mapping(lambda: ProviderRewards()));

        # a mapping between programs and their total stakes
        self._programStakes = mapping(lambda: uint256());

    # solhint-enable func-name-mixedcase

    '''
     * @inheritdoc Upgradeable
    '''
    def version(self) -> (int):
        return 4;

    '''
     * @inheritdoc IStandardRewards
    '''
    def programIds(self) -> (list):
        length = self._nextProgramId - self.INITIAL_PROGRAM_ID;
        return [self.INITIAL_PROGRAM_ID + i for i in range(length)]

    '''
     * @inheritdoc IStandardRewards
    '''
    def programs(self, ids) -> (list):
        length = len(ids);
        return [self._programs[ids[i]] for i in range(length)]

    '''
     * @inheritdoc IStandardRewards
    '''
    def providerProgramIds(self, provider) -> (list):
        return self._programIdsByProvider[provider].values();

    '''
     * @inheritdoc IStandardRewards
    '''
    def programRewards(self, id) -> (Rewards):
        return self._programRewards[id];

    '''
     * @inheritdoc IStandardRewards
    '''
    def providerRewards(self, provider, id) -> (ProviderRewards):
        return self._providerRewards[provider][id];

    '''
     * @inheritdoc IStandardRewards
    '''
    def programStake(self, id) -> (uint):
        return self._programStakes[id].clone();

    '''
     * @inheritdoc IStandardRewards
    '''
    def providerStake(self, provider, id) -> (uint):
        return self._providerRewards[provider][id].stakedAmount.clone();

    '''
     * @inheritdoc IStandardRewards
    '''
    def isProgramActive(self, id) -> (bool):
        return self._isProgramActive(self._programs[id]);

    '''
     * @inheritdoc IStandardRewards
    '''
    def isProgramPaused(self, id) -> (bool):
        return self._isProgramPaused(self._programs[id]);

    '''
     * @inheritdoc IStandardRewards
    '''
    def latestProgramId(self, pool) -> (uint):
        return self._latestProgramIdByPool[pool].clone();

    '''
     * @inheritdoc IStandardRewards
    '''
    def createProgram(self,
        pool,
        totalRewards,
        startTime,
        endTime
    ) -> (uint):
        if (not (self._time() <= startTime and startTime < endTime)):
            revert("InvalidParam");

        # ensure that no program exists for the specific pool
        if (self._isProgramActive(self._programs[self._latestProgramIdByPool[pool]])):
            revert("AlreadyExists");

        poolToken = address(0);
        if (pool is (self._bnt)):
            poolToken = self._bntPoolToken;
        else:
            if (not self._networkSettings.isTokenWhitelisted(pool)):
                revert("NotWhitelisted");

            poolToken = self._network.collectionByPool(pool).poolToken(pool);

        id = self._nextProgramId.clone(); self._nextProgramId += 1;
        rewardRate = totalRewards / (endTime - startTime);

        self._programs[id] = ProgramData({
            'id': id,
            'pool': pool,
            'poolToken': poolToken,
            'rewardsToken': address(self._bnt),
            'isPaused': False,
            'startTime': startTime,
            'endTime': endTime,
            'rewardRate': rewardRate,
            'remainingRewards': rewardRate * (endTime - startTime)
        });

        # set the program as the latest program of the pool
        self._latestProgramIdByPool[pool] = id;

        return id;

    '''
     * @inheritdoc IStandardRewards
    '''
    def terminateProgram(self, id) -> None:
        p = self._programs[id];

        self._verifyProgramActive(p);

        # unset the program from being the latest program of the pool
        del self._latestProgramIdByPool[p.pool];

        endTime = p.endTime;

        # reduce the remaining rewards for the token by the remaining rewards and stop rewards accumulation
        p.remainingRewards -= self._remainingRewards(p);
        p.endTime = self._time();

    '''
     * @inheritdoc IStandardRewards
    '''
    def pauseProgram(self, id, pause) -> None:
        p = self._programs[id];

        self._verifyProgramExists(p);

        prevStatus = p.isPaused;
        if (prevStatus == pause):
            return;

        p.isPaused = pause;

    '''
     * @inheritdoc IStandardRewards
    '''
    def join(self, id, poolTokenAmount) -> None:
        p = self._programs[id];

        self._verifyProgramActiveAndNotPaused(p);

        self._join(self.msg_sender, p, poolTokenAmount, self.msg_sender);

    '''
     * @inheritdoc IStandardRewards
    '''
    def leave(self, id, poolTokenAmount) -> None:
        p = self._programs[id];

        self._verifyProgramExists(p);

        self._leave(self.msg_sender, p, poolTokenAmount);

    '''
     * @inheritdoc IStandardRewards
    '''
    def depositAndJoin(self, id, tokenAmount) -> None:
        p = self._programs[id];

        self._verifyProgramActiveAndNotPaused(p);

        self._depositAndJoin(self.msg_sender, p, tokenAmount);

    '''
     * @inheritdoc IStandardRewards
    '''
    def pendingRewards(self, provider, ids) -> (uint):
        reward = uint256(0);

        for i in range(len(ids)):
            id = ids[i];

            p = self._programs[id];

            self._verifyProgramExists(p);

            newRewardPerToken = self._rewardPerToken(p, self._programRewards[id]);
            providerRewardsData = self._providerRewards[provider][id];

            reward += self._pendingRewards(newRewardPerToken, providerRewardsData);

        return reward;

    '''
     * @inheritdoc IStandardRewards
    '''
    def claimRewards(self, ids) -> (uint):
        reward = self._claimRewards(self.msg_sender, ids, False);

        if (reward == 0):
            return uint256(0);

        self._bntGovernance.mint(self.msg_sender, reward);

        return reward;

    '''
     * @inheritdoc IStandardRewards
    '''
    def stakeRewards(self, ids) -> (StakeAmounts):
        reward = self._claimRewards(self.msg_sender, ids, True);

        if (reward == 0):
            return StakeAmounts({ 'stakedRewardAmount': 0, 'poolTokenAmount': 0 });

        self._bntGovernance.mint(address(self), reward);

        # deposit provider's tokens to the network. Please note, that since we're staking rewards, then the deposit
        # should come from the contract itself, but the pool tokens should be sent to the provider directly
        poolTokenAmount = self._deposit(self.msg_sender, address(self), False, address(self._bnt), reward);

        return StakeAmounts({ 'stakedRewardAmount': reward, 'poolTokenAmount': poolTokenAmount });

    '''
     * @dev adds provider's stake to the program
    '''
    def _join(self,
        provider,
        p,
        poolTokenAmount,
        payer
    ) -> None:
        # take a snapshot of the existing rewards (before increasing the stake)
        data = self._snapshotRewards(p, provider);

        # update both program and provider stakes
        self._programStakes[p.id] += poolTokenAmount;

        prevStake = data.stakedAmount;
        data.stakedAmount = prevStake + poolTokenAmount;

        # unless the payer is the contract itself (in which case, no additional transfer is required), transfer the
        # tokens from the payer (we aren't using safeTransferFrom, since the PoolToken is a fully compliant ERC20 token
        # contract)
        if (payer != address(self)):
            p.poolToken.transferFrom(payer, address(self), poolTokenAmount);

        # add the program to the provider's program list
        self._programIdsByProvider[provider].add(p.id);

    '''
     * @dev removes (some of) provider's stake from the program
    '''
    def _leave(self,
        provider,
        p,
        poolTokenAmount
    ) -> None:
        # take a snapshot of the existing rewards (before decreasing the stake)
        data = self._snapshotRewards(p, provider);

        # update both program and provider stakes
        self._programStakes[p.id] -= poolTokenAmount;

        remainingStake = data.stakedAmount - poolTokenAmount;
        data.stakedAmount = remainingStake;

        # transfer the tokens to the provider (we aren't using safeTransfer, since the PoolToken is a fully
        # compliant ERC20 token contract)
        p.poolToken.transfer(provider, poolTokenAmount);

        # if the provider has removed all of its stake and there are no pending rewards - remove the program from the
        # provider's program list
        if (remainingStake == 0 and data.pendingRewards == 0):
            self._programIdsByProvider[provider].remove(p.id);

    '''
     * @dev deposits provider's stake to the network and returns the received pool token amount
    '''
    def _deposit(self,
        provider,
        payer,
        keepPoolTokens,
        pool,
        tokenAmount
    ) -> (uint):
        poolTokenAmount = uint256(0);
        recipient = address(self) if keepPoolTokens else provider;
        externalPayer = payer != address(self);

        # unless the payer is the contract itself (e.g., during the staking process), in which case the tokens were
        # already claimed and pending in the contract - get the tokens from the provider
        if (externalPayer):
            pool.safeTransferFrom(payer, address(self), tokenAmount);

        pool.ensureApprove(address(self._network), tokenAmount);
        poolTokenAmount = self._network.depositFor(recipient, pool, tokenAmount);

        if (keepPoolTokens and pool is (self._bnt)):
            self._vbnt.safeTransfer(provider, poolTokenAmount);

        return poolTokenAmount;

    '''
     * @dev deposits and adds provider's stake to the program
    '''
    def _depositAndJoin(self,
        provider,
        p,
        tokenAmount
    ) -> None:
        # deposit provider's tokens to the network and let the contract itself to claim the pool tokens so that it can
        # immediately add them to a program
        poolTokenAmount = self._deposit(provider, provider, True, p.pool, tokenAmount);

        # join the existing program, but ensure not to attempt to transfer the tokens from the provider by setting the
        # payer as the contract itself
        self._join(provider, p, poolTokenAmount, address(self));

    '''
     * @dev claims rewards
    '''
    def _claimRewards(self,
        provider,
        ids,
        stake
    ) -> (uint):
        reward = uint256(0);

        for i in range(len(ids)):
            p = self._programs[ids[i]];

            self._verifyProgramNotPaused(p);

            claimData = self._claimRewards_(provider, p);

            if (claimData.reward > 0):
                remainingRewards = p.remainingRewards;

                # a sanity check that the reward amount doesn't exceed the remaining rewards per program
                if (remainingRewards < claimData.reward):
                    revert("RewardsTooHigh");

                # decrease the remaining rewards per program
                self._programs[ids[i]].remainingRewards = remainingRewards - claimData.reward;

                # collect same-reward token rewards
                reward += claimData.reward;

            # if the program is no longer active, has no stake left, and there are no pending rewards - remove the
            # program from the provider's program list
            if (not self._isProgramActive(p) and claimData.stakedAmount == 0):
                self._programIdsByProvider[provider].remove(p.id);

        return reward;

    '''
     * @dev claims rewards and returns the received and the pending reward amounts
    '''
    def _claimRewards_(self, provider, p) -> (ClaimData):
        providerRewardsData = self._snapshotRewards(p, provider);

        reward = providerRewardsData.pendingRewards;

        providerRewardsData.pendingRewards = uint256(0);

        return self.ClaimData({ 'reward': reward, 'stakedAmount': providerRewardsData.stakedAmount });

    '''
     * @dev returns whether the specified program is active
    '''
    def _isProgramActive(self, p) -> (bool):
        currTime = self._time();

        return \
            self._programExists(p) and \
            p.startTime <= currTime and \
            currTime <= p.endTime and \
            self._latestProgramIdByPool[p.pool] == p.id;

    '''
     * @dev returns whether the specified program is paused
    '''
    def _isProgramPaused(self, p) -> (bool):
        return p.isPaused;

    '''
     * @dev returns whether or not a given program exists
    '''
    def _programExists(self, p) -> (bool):
        return address(p.pool) != address(0);

    '''
     * @dev verifies that a program exists
    '''
    def _verifyProgramExists(self, p) -> None:
        if (not self._programExists(p)):
            revert("DoesNotExist");

    '''
     * @dev verifies that a program exists, and active
    '''
    def _verifyProgramActive(self, p) -> None:
        self._verifyProgramExists(p);

        if (not self._isProgramActive(p)):
            revert("ProgramInactive");

    '''
     * @dev verifies that a program is not paused
    '''
    def _verifyProgramNotPaused(self, p) -> None:
        self._verifyProgramExists(p);

        if (p.isPaused):
            revert("ProgramSuspended");

    '''
     * @dev verifies that a program exists, active, and not paused
    '''
    def _verifyProgramActiveAndNotPaused(self, p) -> None:
        self._verifyProgramActive(p);
        self._verifyProgramNotPaused(p);

    '''
     * @dev returns the remaining rewards of given program
    '''
    def _remainingRewards(self, p) -> (uint):
        currTime = self._time();

        return p.rewardRate * (p.endTime - currTime) if p.endTime > currTime else uint256(0);

    '''
     * @dev updates program and provider's rewards
    '''
    def _snapshotRewards(self, p, provider) -> (ProviderRewards):
        rewards = self._programRewards[p.id];

        newRewardPerToken = self._rewardPerToken(p, rewards);
        if (newRewardPerToken != rewards.rewardPerToken):
            rewards.rewardPerToken = newRewardPerToken;

        newUpdateTime = uint32(Math.min(self._time(), p.endTime));
        if (rewards.lastUpdateTime < newUpdateTime):
            rewards.lastUpdateTime = newUpdateTime;

        providerRewardsData = self._providerRewards[provider][p.id];

        newPendingRewards = self._pendingRewards(newRewardPerToken, providerRewardsData);
        if (newPendingRewards != 0):
            providerRewardsData.pendingRewards = newPendingRewards;

        providerRewardsData.rewardPerTokenPaid = newRewardPerToken;

        return providerRewardsData;

    '''
     * @dev calculates current reward per-token amount
    '''
    def _rewardPerToken(self, p, rewards) -> (uint):
        currTime = self._time();
        if (currTime < p.startTime):
            return uint256(0);

        totalStaked = self._programStakes[p.id];
        if (totalStaked == 0):
            return rewards.rewardPerToken;

        stakingEndTime = Math.min(currTime, p.endTime);
        stakingStartTime = Math.max(p.startTime, rewards.lastUpdateTime);

        return \
            rewards.rewardPerToken + \
            (((stakingEndTime - stakingStartTime) * p.rewardRate * self.REWARD_RATE_FACTOR) / totalStaked);

    '''
     * @dev calculates provider's pending rewards
    '''
    def _pendingRewards(self, updatedRewardPerToken, providerRewardsData) -> (uint):
        return \
            providerRewardsData.pendingRewards + \
            (providerRewardsData.stakedAmount * (updatedRewardPerToken - providerRewardsData.rewardPerTokenPaid)) / \
            self.REWARD_RATE_FACTOR;

    '''
     * @dev returns whether the specified array has duplicates
    '''
    def _isArrayUnique(self, ids) -> (bool):
        for i in range(len(ids)):
            for j in range(i + 1, len(ids)):
                if (ids[i] == ids[j]):
                    return False;

        return True;



from bancor_research import Decimal

class uint:
    sizes = [(n + 1) * 8 for n in range(32)]

    def __init__(self, size, other):
        assert size in uint.sizes
        self.size = size
        self.data = uint._data(other)

    def clone(self):
        return uint(self.size, self.data)

    def __add__(self, other):
        return self._new(other, Decimal.__add__)

    def __sub__(self, other):
        return self._new(other, Decimal.__sub__)

    def __mul__(self, other):
        return self._new(other, Decimal.__mul__)

    def __truediv__(self, other):
        return self._new(other, Decimal.__truediv__)

    def __mod__(self, other):
        return self._new(other, Decimal.__mod__)

    def __pow__(self, other):
        return self._new(other, Decimal.__pow__)

    def __iadd__(self, other):
        return self._set(self + other)

    def __isub__(self, other):
        return self._set(self - other)

    def __imul__(self, other):
        return self._set(self * other)

    def __itruediv__(self, other):
        return self._set(self / other)

    def __imod__(self, other):
        return self._set(self % other)

    def __ipow__(self, other):
        return self._set(self ** other)

    def __lt__(self, other):
        return self.data < uint._data(other)

    def __le__(self, other):
        return self.data <= uint._data(other)

    def __eq__(self, other):
        return self.data == uint._data(other)

    def __ne__(self, other):
        return self.data != uint._data(other)

    def __gt__(self, other):
        return self.data > uint._data(other)

    def __ge__(self, other):
        return self.data >= uint._data(other)

    def __str__(self):
        return str(self.data)

    def __hash__(self):
        return hash(self.data)

    def _set(self, other):
        assert self.size >= other.size
        self.data = other.data
        return self

    def _new(self, other, op):
        data = op(self.data, uint._data(other))
        size = max(self.size, uint._size(other))
        return uint(size, data)

    @staticmethod
    def _data(other):
        return other.data if type(other) is uint else Decimal(other)

    @staticmethod
    def _size(other):
        return other.size if type(other) is uint else 0

class unchecked:

    @staticmethod
    def begin():
        pass

    @staticmethod
    def end():
        pass



from bancor_research.bancor_emulator.solidity import uint, uint32, block

'''
 * @dev this contract abstracts the block number in order to allow for more flexible control in tests
'''
class BlockNumber:
    '''
     * @dev returns the current block-number
    '''
    def _blockNumber(self) -> (uint):
        return uint32(block.number);



from bancor_research.bancor_emulator.ERC20 import ERC20

'''
 * @dev ReserveToken represents a token with dynamic supply.
 * The owner of the token can mint/burn tokens to/from any account.
 *
'''
class ReserveToken(ERC20):
    '''
     * @dev initializes a new ReserveToken instance
     *
     * @param _name       token name
     * @param _symbol     token short symbol, minimum 1 character
     * @param _decimals   for display purposes only
    '''
    def __init__(self,
        _name,
        _symbol,
        _decimals
    ) -> None:
        ERC20.__init__(self, _name, _symbol)

        self._decimals = _decimals;

    '''
     * @dev increases the token supply and sends the new tokens to the given account
     * can only be called by the contract owner
     *
     * @param _to      account to receive the new amount
     * @param _amount  amount to increase the supply by
    '''
    def issue(self, _to, _amount):
        self._mint(_to, _amount);

    '''
     * @dev removes tokens from the given account and decreases the token supply
     * can only be called by the contract owner
     *
     * @param _from    account to remove the amount from
     * @param _amount  amount to decrease the supply by
    '''
    def destroy(self, _from, _amount):
        self._burn(_from, _amount);

    def decimals(self):
        return self._decimals;



# coding=utf-8
# --------------------------------------------------------------------------------------------------------------------
# Licensed under the MIT LICENSE. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------------------------------
"""Main BancorDapp class and simulator module interface."""

import cloudpickle, pandas

from bancor_research.bancor_simulator.v3.spec.actions import *
from bancor_research.bancor_simulator.v3.spec.rewards import *
from bancor_research.bancor_simulator.v3.spec.state import *

from bancor_research import DataFrame, read_price_feeds


def to_decimal(percent: str):
    try:
        return Decimal(percent[:-1]) / 100
    except TypeError:
        return percent


def to_user_amount(state: State, tkn_name: str, user_name: str, amount: str):
    if amount.endswith("%"):
        return get_user_balance(state, user_name, tkn_name) * to_decimal(amount)
    return Decimal(amount)


class BancorDapp:
    """Main BancorDapp class and simulator module interface."""

    """
    Args:
        timestamp (integer, default = 0): The Ethereum block number to begin the simulation with
        bnt_min_liquidity (numeric string, default = 10000): The minimum BNT liquidity needed to bootstrap a pool
        withdrawal_fee (percentage string, default = 0.25%): The global exit (withdrawal) fee
        coolown_time (integer, default = 7 days): The cooldown period in seconds
        network_fee (percentage string, default = 20%): The global network fee
        whitelisted_tokens (dictionary): The configuration of each pool (default = ["eth", "link", "tkn", "wbtc"]):
        - trading_fee (percentage string, default = 1%): The trading fee of the pool
        - bnt_funding_limit (numeric string, default = 1000000): The BNT funding limit of the pool
        price_feeds_path (string): The path to a file containing price feeds
        price_feeds (dictionary): The price feeds to use instead of the file
    """

    def __init__(
        self,
        timestamp: int = DEFAULT.TIMESTAMP,
        bnt_min_liquidity: str = DEFAULT.BNT_MIN_LIQUIDITY,
        withdrawal_fee: str = DEFAULT.WITHDRAWAL_FEE,
        cooldown_time: int = DEFAULT.COOLDOWN_TIME,
        network_fee: str = DEFAULT.NETWORK_FEE,
        whitelisted_tokens=DEFAULT.WHITELIST,
        price_feeds_path: str = DEFAULT.PRICE_FEEDS_PATH,
        price_feeds: PandasDataFrame = DEFAULT.PRICE_FEEDS,
        log_state: bool = True,
    ):

        transaction_id = 0

        self.json_data = None
        self.transaction_id = transaction_id

        self.price_feeds = (
            price_feeds
            if price_feeds is not None
            else read_price_feeds(price_feeds_path)
        )

        for tkn_name in whitelisted_tokens:
            assert tkn_name in price_feeds.columns, (
                f"Whitelisted token `{tkn_name}` not found in price feed. "
                f"Add `{tkn_name}` to the price feed, "
                f"or remove `{tkn_name}` from the whitelisted tokens."
            )

        whitelisted_tokens = {
            k: {
                "decimals": v["decimals"],
                "trading_fee": to_decimal(v["trading_fee"]),
                "bnt_funding_limit": Decimal(v["bnt_funding_limit"]),
                "ep_vault_balance": Decimal(v["ep_vault_balance"]),
            }
            for k, v in whitelisted_tokens.items()
        }

        state = State(
            transaction_id=transaction_id,
            timestamp=timestamp,
            whitelisted_tokens=whitelisted_tokens,
            price_feeds=price_feeds,
        )

        state = init_protocol(
            state=state,
            whitelisted_tokens=whitelisted_tokens,
            usernames=[],
            cooldown_time=cooldown_time,
            network_fee=to_decimal(network_fee),
            bnt_min_liquidity=Decimal(bnt_min_liquidity),
            withdrawal_fee=to_decimal(withdrawal_fee),
        )

        state.json_export = {"users": [], "operations": []}
        self._backup_states = {}
        self._global_state = state
        self.history = []
        self.log_state = log_state

    @property
    def backup_states(self):
        return [i for i in self._backup_states]

    @property
    def global_state(self):
        return self._global_state

    @global_state.setter
    def global_state(self, value):
        self._global_state = value

    def copy_state(self, copy_type: str, state: State = None, timestamp: int = 0):
        """
        Saves a backup of the current global state to revert_state back to if desired.
        """
        assert copy_type in ["initial", "end"], "copy_type must be 'initial' or 'end'"

        if state is None:
            s = self.global_state.copy()
        else:
            s = state.copy()

        if timestamp is None:
            ts = s.timestamp
        else:
            ts = timestamp
            s.timestamp = timestamp
        self._backup_states[f"{copy_type}_{ts}"] = s.copy()
        return s

    def update_state(self, state: State, timestamp: int = 0):
        """
        Updates the global state at the end of each action.
        """
        self.global_state = self.copy_state("end", state, timestamp)

    def next_transaction(self, state: State, timestamp: int = 0):
        """
        Increments a new id and state for each action
        """
        if self.log_state:
            self.update_state(state, timestamp)
        self.transaction_id += 1

    def get_state(
        self, copy_type: str = "initial", state: State = None, timestamp: int = 0
    ):
        """
        Creates a copy of the global state which will modified during a new action.
        """
        if self.log_state:
            return self.copy_state(
                copy_type=copy_type, state=state, timestamp=timestamp
            )
        else:
            self.global_state.timestamp = timestamp
            return self.global_state

    def revert_state(self, timestamp):
        """
        Reverts the state of the bancor network to a previously saved state.
        """
        self.global_state = self._backup_states[timestamp]

    def show_history(self):
        """
        Displays the history of the bancor network in a dataframe.
        """
        return pandas.concat(
            [
                self.global_state.history[i]
                for i in range(len(self.global_state.history))
            ]
        )

    def deposit(
        self,
        tkn_name: str,
        tkn_amt: str,
        user_name: str,
        timestamp: int = 0,
        action_name="deposit",
    ):
        """
        Top level logic for deposit actions.
        """
        state = self.get_state(copy_type="initial", timestamp=timestamp)
        state, tkn_name, user_name = validate_input(
            state, tkn_name, user_name, timestamp
        )
        tkn_amt = to_user_amount(state, tkn_name, user_name, tkn_amt)
        if tkn_name == "bnt":
            state = deposit_bnt(
                state=state, tkn_name=tkn_name, tkn_amt=tkn_amt, user_name=user_name
            )
        else:
            state = handle_vandalism_attack(state, tkn_name)
            state = deposit_tkn(
                state=state, tkn_name=tkn_name, tkn_amt=tkn_amt, user_name=user_name
            )
        state = handle_logging(
            tkn_name=tkn_name,
            tkn_amt=tkn_amt,
            action_name=action_name,
            user_name=user_name,
            transaction_id=self.transaction_id,
            state=state,
        )
        self.next_transaction(state)

    def trade(
        self,
        tkn_amt: str,
        source_token: str,
        target_token: str,
        user_name: str,
        timestamp: int,
        transaction_type: str = "trade",
    ):
        """
        Main logic for trade actions.
        """
        state = self.get_state(copy_type="initial", timestamp=timestamp)
        state, source_token, user_name = validate_input(
            state, source_token, user_name, timestamp
        )
        state, target_token, user_name = validate_input(
            state, target_token, user_name, timestamp
        )
        tkn_amt = to_user_amount(state, source_token, user_name, tkn_amt)
        state = process_trade(
            state, tkn_amt, source_token, target_token, user_name, timestamp
        )
        state = handle_logging(
            source_token + "->" + target_token,
            tkn_amt,
            transaction_type,
            user_name,
            self.transaction_id,
            state,
        )
        self.next_transaction(state)

    def begin_cooldown_by_rtkn(
        self,
        tkn_amt: str,
        tkn_name: str,
        user_name: str,
        timestamp: int = 0,
        action_name: str = "begin cooldown by reserve token amount",
    ):
        """
        Begin the withdrawal cooldown operation by specifying the output amount of reserve tokens.
        """
        state = self.get_state(copy_type="initial", timestamp=timestamp)
        state, tkn_name, user_name = validate_input(
            state, tkn_name, user_name, timestamp
        )
        tkn_amt = to_user_amount(state, tkn_name, user_name, tkn_amt)
        id_number = begin_withdrawal_cooldown_by_rtkn(
            state, tkn_amt, tkn_name, user_name
        )
        state = handle_logging(
            tkn_name, tkn_amt, action_name, user_name, self.transaction_id, state
        )
        self.next_transaction(state)

        return id_number

    def begin_cooldown_by_ptkn(
        self,
        tkn_amt: str,
        tkn_name: str,
        user_name: str,
        timestamp: int = 0,
        action_name: str = "begin cooldown by pool token amount",
    ):
        """
        Begin the withdrawal cooldown operation by specifying the intput amount of pool tokens.
        """
        state = self.get_state(copy_type="initial", timestamp=timestamp)
        state, tkn_name, user_name = validate_input(
            state, tkn_name, user_name, timestamp
        )
        tkn_amt = to_user_amount(state, "bn" + tkn_name, user_name, tkn_amt)
        id_number = begin_withdrawal_cooldown_by_ptkn(
            state, tkn_amt, tkn_name, user_name
        )
        self.next_transaction(state)
        handle_logging(
            tkn_name, tkn_amt, action_name, user_name, self.transaction_id, state
        )
        return id_number

    def withdraw(
        self,
        user_name: str,
        id_number: int,
        timestamp: int = 0,
        transaction_type: str = "withdraw",
    ):
        """
        Main withdrawal logic based on the withdraw algorithm of the BIP15 spec.
        """
        state = self.get_state(copy_type="initial", timestamp=timestamp)
        tkn_name = state.users[user_name].pending_withdrawals[id_number].tkn_name
        state = process_withdrawal(state, user_name, id_number, timestamp)
        state = handle_logging(
            tkn_name,
            Decimal(0),
            transaction_type,
            user_name,
            self.transaction_id,
            state,
        )
        self.next_transaction(state)

    def enable_trading(
        self,
        tkn_name: str,
        tkn_price: str = "",
        bnt_price: str = "",
        timestamp: int = 0,
        transaction_type: str = "enableTrading",
    ) -> None:
        """
        DAO msig initilizes new tokens to allow trading once specified conditions are met.
        """
        state = self.get_state(copy_type="initial", timestamp=timestamp)
        state, tkn_name, user_name = validate_input(state, tkn_name, "", timestamp)
        if tkn_price:
            state.price_feeds.at[state.timestamp, tkn_name] = tkn_price
        if bnt_price:
            state.price_feeds.at[state.timestamp, "bnt"] = bnt_price
        state = enable_trading(state, tkn_name)
        state = handle_logging(
            tkn_name,
            Decimal("0"),
            transaction_type,
            user_name,
            state.transaction_id,
            state,
        )
        self.next_transaction(state)

    def describe(self, decimals: int = -1):
        """
        Describes the state ledger in a format similar to BIP15 documentation.
        """
        table = {}

        state = self.global_state

        reserve_tokens = ["bnt"] + [tkn for tkn in state.whitelisted_tokens]
        pool_tokens = ["bnbnt"] + ["bn" + tkn for tkn in state.whitelisted_tokens]

        # Iterate all tokens
        for tkn_name in reserve_tokens + pool_tokens + ["vbnt"]:
            table[tkn_name] = {}
            for account in [user for user in state.users if user != protocol_user]:
                table[tkn_name][tuple([1, "Account", account])] = (
                    state.users[account].wallet[tkn_name].balance
                )

        # Iterate all reserve tokens except bnt
        for tkn_name in [tkn_name for tkn_name in reserve_tokens if tkn_name != "bnt"]:
            table[tkn_name][tuple([2, "Pool", "a: TKN Staked Balance"])] = state.tokens[
                tkn_name
            ].staking_ledger.balance
            table[tkn_name][
                tuple([2, "Pool", "b: TKN Trading Liquidity"])
            ] = state.tokens[tkn_name].tkn_trading_liquidity.balance
            table[tkn_name][
                tuple([2, "Pool", "c: BNT Trading Liquidity"])
            ] = state.tokens[tkn_name].bnt_trading_liquidity.balance
            table[tkn_name][
                tuple([2, "Pool", "d: BNT Current Funding"])
            ] = state.tokens[tkn_name].bnt_funding_amt.balance
            table[tkn_name][tuple([2, "Pool", "e: Spot Rate"])] = state.tokens[
                tkn_name
            ].spot_rate
            table[tkn_name][tuple([2, "Pool", "f: Average Rate"])] = state.tokens[
                tkn_name
            ].ema_rate
            table[tkn_name][
                tuple([2, "Pool", "g: Average Inverse Rate"])
            ] = state.tokens[tkn_name].inv_ema_rate

        # Iterate all reserve tokens
        for tkn_name in reserve_tokens:
            table[tkn_name][tuple([3, "Network", "Master Vault"])] = state.tokens[
                tkn_name
            ].master_vault.balance
            table[tkn_name][tuple([3, "Network", "Protection Vault"])] = state.tokens[
                tkn_name
            ].external_protection_vault.balance
            table["bn" + tkn_name][
                tuple([3, "Network", "Rewards Vault"])
            ] = state.tokens[tkn_name].external_rewards_vault.balance
            table["bn" + tkn_name][
                tuple([3, "Network", "Protocol Equity"])
            ] = state.tokens[tkn_name].protocol_wallet_pooltokens.balance

        # Iterate only bnt
        table["bnt"][tuple([3, "Network", "BNT Staked Balance"])] = state.tokens[
            "bnt"
        ].staking_ledger.balance
        table["bnt"][tuple([3, "Network", "Vortex"])] = state.tokens[
            "bnt"
        ].vortex_ledger.balance

        df = DataFrame(table).fillna(0).sort_index()
        return df.applymap(lambda x: round(x, decimals)) if decimals >= 0 else df

    def export(self):
        """
        Exports transaction history record
        """
        return pandas.concat(self.global_state.history)

    def whitelist_token(self, tkn_name: str, timestamp: int = 0):
        """
        Creates a new whitelisted token with initialized starting balances
        """
        state = self.get_state(copy_type="initial", timestamp=timestamp)
        tkn_name = tkn_name.lower()
        state.price_feeds[tkn_name] = state.price_feeds["bnt"].values

        state.create_whitelisted_tkn(tkn_name)
        handle_whitelisting_tokens(state)

        self.next_transaction(state)
        handle_logging(
            tkn_name,
            Decimal("0"),
            f"whitelist_{tkn_name}",
            "NA",
            state.transaction_id,
            state,
        )

    def create_user(self, user_name: str, timestamp: int = 0):
        """
        Creates a new user with a valid wallet
        """
        state = self.get_state(copy_type="initial", timestamp=timestamp)
        state.create_user(user_name)
        self.next_transaction(state)
        handle_logging(
            "NA", Decimal("0"), f"create_{user_name}", "NA", state.transaction_id, state
        )

    def process_ac_rewards_program(
        self,
        tkn_name: str,
        timestamp: int = 0,
        transaction_type: str = "process autocompounding rewards program",
    ):
        """
        Processes autocompounding rewards program.
        """
        state = self.get_state(copy_type="initial", timestamp=timestamp)
        state, tkn_name, user_name = validate_input(state, tkn_name, "", timestamp)
        state = process_ac_rewards_program(
            state=state, tkn_name=tkn_name, timestamp=timestamp
        )
        self.next_transaction(state)
        handle_logging(
            tkn_name=tkn_name,
            user_name=user_name,
            tkn_amt=Decimal(0),
            action_name=transaction_type,
            transaction_id=self.transaction_id,
            state=state,
        )

    def load_json_simulation(self, path, tkn_name="tkn", timestamp: int = 0):
        """
        Loads a JSON file containing simulation modules to run and report on.
        """
        tkn_name = tkn_name.lower()
        state = self.get_state(copy_type="initial", timestamp=timestamp)
        json_data = BancorDapp.load_json(path)
        state = setup_json_simulation(state, json_data, tkn_name)
        self.next_transaction(state)

    def create_flat_ac_rewards_program(
        self,
        tkn_name: str,
        user_name: str,
        total_rewards: str,
        start_time: int,
        total_duration: int,
        timestamp: int = 0,
        transaction_type: str = "create flat autocompounding rewards program",
    ):
        self._create_ac_rewards_program(
            tkn_name=tkn_name,
            user_name=user_name,
            total_rewards=total_rewards,
            distribution_type="flat",
            start_time=start_time,
            total_duration=total_duration,
            timestamp=timestamp,
            transaction_type=transaction_type,
        )

    def create_exp_ac_rewards_program(
        self,
        tkn_name: str,
        user_name: str,
        total_rewards: str,
        start_time: int,
        half_life: int,
        timestamp: int = 0,
        transaction_type: str = "create exp autocompounding rewards program",
    ):
        self._create_ac_rewards_program(
            tkn_name=tkn_name,
            user_name=user_name,
            total_rewards=total_rewards,
            distribution_type="exp",
            start_time=start_time,
            half_life=half_life,
            timestamp=timestamp,
            transaction_type=transaction_type,
        )

    def _create_ac_rewards_program(
        self,
        tkn_name: str,
        user_name: str,
        total_rewards: str,
        distribution_type: str,
        start_time: int,
        total_duration: int = 0,
        half_life: int = 0,
        timestamp: int = 0,
        transaction_type: str = "",
    ):
        """
        Creates autocompounding rewards program.
        """
        state = self.get_state(copy_type="initial", timestamp=timestamp)
        state, tkn_name, user_name = validate_input(
            state, tkn_name, user_name, timestamp
        )

        program_id = state.autocompounding_programs_count + 1

        is_active = timestamp >= start_time

        total_rewards = to_user_amount(state, tkn_name, user_name, total_rewards)

        ptkn_amt = compute_ptkn_amt(state, tkn_name, total_rewards)
        state.decrease_user_balance(user_name, "bn" + tkn_name, ptkn_amt)
        state.increase_external_rewards_vault_balance(tkn_name, ptkn_amt)

        # Add the program to the rest.
        state.autocompounding_reward_programs[tkn_name] = AutocompoundingProgram(
            id=program_id,
            created_at=timestamp,
            tkn_name=tkn_name,
            distribution_type=distribution_type,
            start_time=start_time,
            total_duration=total_duration,
            half_life=half_life,
            total_rewards=Token(balance=total_rewards),
            remaining_rewards=Token(balance=total_rewards),
            is_active=is_active,
        )

        self.next_transaction(state)
        handle_logging(
            tkn_name=tkn_name,
            user_name=user_name,
            tkn_amt=total_rewards,
            action_name=transaction_type,
            transaction_id=self.transaction_id,
            state=state,
        )

    def terminate_ac_rewards_program(
        self,
        tkn_name: str,
        timestamp: int = 0,
        transaction_type: str = "terminate autocompounding rewards program",
    ):
        """
        Terminates autocompounding rewards program.
        """
        state = self.get_state(copy_type="initial", timestamp=timestamp)
        state, tkn_name, user_name = validate_input(state, tkn_name, "", timestamp)
        state = terminate_ac_rewards_program(
            state=state, tkn_name=tkn_name, timestamp=timestamp
        )
        self.next_transaction(state)
        handle_logging(
            tkn_name=tkn_name,
            user_name=user_name,
            tkn_amt=Decimal(0),
            action_name=transaction_type,
            transaction_id=self.transaction_id,
            state=state,
        )

    def burn_pool_tokens(
        self,
        tkn_name: str,
        tkn_amt: str,
        user_name: str,
        timestamp: int = 0,
        transaction_type: str = "burnPoolTokenTKN",
    ):
        """
        Used for testing vandalism attack.
        """
        state = self.get_state(copy_type="initial", timestamp=timestamp)
        state, tkn_name, user_name = validate_input(
            state, tkn_name, user_name, timestamp
        )
        tkn_amt = to_user_amount(state, "bn" + tkn_name, user_name, tkn_amt)
        if tkn_name != "bnt":
            state.decrease_pooltoken_balance(tkn_name, tkn_amt)
            state.decrease_user_balance(
                user_name, get_pooltoken_name(tkn_name), tkn_amt
            )
            self.next_transaction(state)
            handle_logging(
                tkn_name,
                tkn_amt,
                transaction_type,
                user_name,
                self.transaction_id,
                state,
            )

    def create_standard_rewards_program(
        self,
        tkn_name: str,
        rewards_amt: str,
        start_time: int,
        end_time: int,
        timestamp: int,
        transaction_type="create_standard_rewards_program",
    ):
        """
        Create a standard rewards program for a given token.
        """
        state = self.get_state(copy_type="initial", timestamp=timestamp)
        state, tkn_name, user_name = validate_input(state, tkn_name, "", timestamp)
        tkn_amt = Decimal(rewards_amt)
        state, id = create_standard_reward_program(
            state=state,
            tkn_name=tkn_name,
            total_rewards=tkn_amt,
            start_time=start_time,
            end_time=end_time,
            timestamp=timestamp,
        )
        self.next_transaction(state)
        handle_logging(
            tkn_name, tkn_amt, transaction_type, user_name, self.transaction_id, state
        )
        return id

    def join_standard_rewards_program(
        self,
        tkn_name: str,
        tkn_amt: str,
        user_name: str,
        program_id: int,
        timestamp: int = 0,
        transaction_type="join_standard_rewards_program",
    ):
        """
        Join the standard rewards program for a given user.
        """
        state = self.get_state(copy_type="initial", timestamp=timestamp)
        state, tkn_name, user_name = validate_input(
            state, tkn_name, user_name, timestamp
        )
        tkn_amt = to_user_amount(state, "bn" + tkn_name, user_name, tkn_amt)
        state = join_standard_reward_program(
            state=state,
            user_name=user_name,
            id=program_id,
            pool_token_amount=tkn_amt,
            timestamp=timestamp,
        )
        self.next_transaction(state)
        handle_logging(
            tkn_name, tkn_amt, transaction_type, user_name, self.transaction_id, state
        )

    def leave_standard_rewards_program(
        self,
        tkn_name: str,
        tkn_amt: str,
        user_name: str,
        program_id: int,
        timestamp: int = 0,
        transaction_type="leave_standard_rewards_program",
    ):
        """
        Leave the standard rewards program for a given user.
        """
        state = self.get_state(copy_type="initial", timestamp=timestamp)
        state, tkn_name, user_name = validate_input(
            state, tkn_name, user_name, timestamp
        )
        tkn_amt = to_user_amount(state, "bn" + tkn_name, user_name, tkn_amt)
        state = leave_standard_reward_program(
            state=state,
            user_name=user_name,
            id=program_id,
            pool_token_amount=tkn_amt,
            timestamp=timestamp,
        )
        self.next_transaction(state)
        handle_logging(
            tkn_name, tkn_amt, transaction_type, user_name, self.transaction_id, state
        )

    def claim_standard_rewards(
        self,
        user_name: str,
        program_ids: list,
        timestamp: int = 0,
        transaction_type: str = "claim_standard_rewards",
    ):
        """
        Claim standard rewards for a given reward program and user.
        """
        state = self.get_state(copy_type="initial", timestamp=timestamp)
        state = claim_standard_rewards(
            state=state,
            user_name=user_name,
            ids=program_ids,
            timestamp=timestamp,
        )
        self.next_transaction(state)
        handle_logging(
            "",
            Decimal(0),
            transaction_type,
            user_name,
            self.transaction_id,
            state,
        )

    def set_user_balance(
        self,
        user_name: str,
        tkn_name: str,
        tkn_amt: str,
        timestamp: int = 0,
        transaction_type: str = "set user balance",
    ):
        """
        Sets user balance at the network interface level for convenience.
        """
        state = self.get_state(copy_type="initial", timestamp=timestamp)
        state, tkn_name, user_name = validate_input(
            state, tkn_name, user_name, timestamp
        )
        tkn_amt = Decimal(tkn_amt)
        state.set_user_balance(user_name, tkn_name, tkn_amt)
        self.next_transaction(state)
        handle_logging(
            tkn_name, tkn_amt, transaction_type, user_name, self.transaction_id, state
        )

    def step(self):
        self.global_state.step()

    def set_state(self, state: State):
        self.global_state = state

    def set_trading_fee(
        self,
        tkn_name: str,
        percent: str,
        timestamp: int = 0,
        transaction_type: str = "set trading fee",
    ):
        """
        Sets the system trading fee at the network interface level for convenience.
        """
        state = self.get_state(copy_type="initial", timestamp=timestamp)
        state, tkn_name, user_name = validate_input(state, tkn_name, "", timestamp)
        value = to_decimal(percent)
        state.set_trading_fee(tkn_name, value)
        self.next_transaction(state)
        handle_logging(
            tkn_name, value, transaction_type, user_name, self.transaction_id, state
        )
        return self

    def set_network_fee(
        self,
        tkn_name: str,
        percent: str,
        timestamp: int = 0,
        transaction_type: str = "set network fee",
    ):
        """
        Sets the system network fee at the network interface level for convenience.
        """
        state = self.get_state(copy_type="initial", timestamp=timestamp)
        state, tkn_name, user_name = validate_input(state, tkn_name, "", timestamp)
        value = to_decimal(percent)
        state.set_network_fee(tkn_name, value)
        self.next_transaction(state)
        handle_logging(
            tkn_name, value, transaction_type, user_name, self.transaction_id, state
        )
        return self

    def set_withdrawal_fee(
        self,
        tkn_name: str,
        percent: str,
        timestamp: int = 0,
        transaction_type: str = "set withdrawal fee",
    ):
        """
        Sets the system withdrawal fee at the network interface level for convenience.
        """
        state = self.get_state(copy_type="initial", timestamp=timestamp)
        state, tkn_name, user_name = validate_input(state, tkn_name, "", timestamp)
        value = to_decimal(percent)
        state.set_withdrawal_fee(tkn_name, value)
        self.next_transaction(state)
        handle_logging(
            tkn_name, value, transaction_type, user_name, self.transaction_id, state
        )
        return self

    def set_bnt_funding_limit(
        self,
        tkn_name: str,
        amount: str,
        timestamp: int = 0,
        transaction_type: str = "set bnt funding limit",
    ):
        """
        Sets the system withdrawal fee at the network interface level for convenience.
        """
        state = self.get_state(copy_type="initial", timestamp=timestamp)
        state, tkn_name, user_name = validate_input(state, tkn_name, "", timestamp)
        value = to_decimal(amount)
        state.set_bnt_funding_limit(tkn_name, value)
        self.next_transaction(state)
        handle_logging(
            tkn_name, value, transaction_type, user_name, self.transaction_id, state
        )
        return self

    def save(self, file_path, pickle_protocol=cloudpickle.DEFAULT_PROTOCOL):
        """
        Saves state at file path.
        """
        with open(file_path, "wb") as f:
            cloudpickle.dump(self, f, protocol=pickle_protocol)



from bancor_research.bancor_emulator.solidity import uint, uint256, unchecked, mulmod, revert
from bancor_research.bancor_emulator.utils import library, parse

from bancor_research.bancor_emulator.Math import Math
from bancor_research.bancor_emulator.Fraction import Fraction256
from bancor_research.bancor_emulator.Constants import PPM_RESOLUTION

ONE = 0x80000000000000000000000000000000;
LN2 = 0x58b90bfbe8e7bcd5e4f1d9cc01f97b57;

class Uint512:
    def __init__(self, x = None) -> None:
        self.hi = parse(uint256, x, 'hi'); # 256 most significant bits
        self.lo = parse(uint256, x, 'lo'); # 256 least significant bits

class Sint256:
    def __init__(self, x = None) -> None:
        self.value = parse(uint256, x, 'value');
        self.isNeg = parse(bool, x, 'isNeg');

'''
 * @dev this library provides a set of complex math operations
'''
class MathEx:
    exp2 = None
    truncatedFraction = None
    weightedAverage = None
    isInRange = None
    toPos256 = None
    toNeg256 = None
    mulDivF = None
    mulDivC = None
    subMax0 = None
    gt512 = None
    lt512 = None
    gte512 = None
    lte512 = None
    mul512 = None

'''
    * @dev returns `2 ^ f` by calculating `e ^ (f * ln(2))`, where `e` is Euler's number:
    * - Rewrite the input as a sum of binary exponents and a single residual r, as small as possible
    * - The exponentiation of each binary exponent is given (pre-calculated)
    * - The exponentiation of r is calculated via Taylor series for e^x, where x = r
    * - The exponentiation of the input is calculated by multiplying the intermediate results above
    * - For example: e^5.521692859 = e^(4 + 1 + 0.5 + 0.021692859) = e^4 * e^1 * e^0.5 * e^0.021692859
'''
def exp2(f: Fraction256) -> (Fraction256):
    unchecked.begin()

    x = mulDivF(LN2, f.n, f.d);
    y = uint256();
    z = uint256();
    n = uint256();

    if (x >= (ONE << 4)):
        revert("Overflow");

    z = y = x % (ONE >> 3); # get the input modulo 2^(-3)
    z = (z * y) / ONE;
    n += z * 0x10e1b3be415a0000; # add y^02 * (20! / 02!)
    z = (z * y) / ONE;
    n += z * 0x05a0913f6b1e0000; # add y^03 * (20! / 03!)
    z = (z * y) / ONE;
    n += z * 0x0168244fdac78000; # add y^04 * (20! / 04!)
    z = (z * y) / ONE;
    n += z * 0x004807432bc18000; # add y^05 * (20! / 05!)
    z = (z * y) / ONE;
    n += z * 0x000c0135dca04000; # add y^06 * (20! / 06!)
    z = (z * y) / ONE;
    n += z * 0x0001b707b1cdc000; # add y^07 * (20! / 07!)
    z = (z * y) / ONE;
    n += z * 0x000036e0f639b800; # add y^08 * (20! / 08!)
    z = (z * y) / ONE;
    n += z * 0x00000618fee9f800; # add y^09 * (20! / 09!)
    z = (z * y) / ONE;
    n += z * 0x0000009c197dcc00; # add y^10 * (20! / 10!)
    z = (z * y) / ONE;
    n += z * 0x0000000e30dce400; # add y^11 * (20! / 11!)
    z = (z * y) / ONE;
    n += z * 0x000000012ebd1300; # add y^12 * (20! / 12!)
    z = (z * y) / ONE;
    n += z * 0x0000000017499f00; # add y^13 * (20! / 13!)
    z = (z * y) / ONE;
    n += z * 0x0000000001a9d480; # add y^14 * (20! / 14!)
    z = (z * y) / ONE;
    n += z * 0x00000000001c6380; # add y^15 * (20! / 15!)
    z = (z * y) / ONE;
    n += z * 0x000000000001c638; # add y^16 * (20! / 16!)
    z = (z * y) / ONE;
    n += z * 0x0000000000001ab8; # add y^17 * (20! / 17!)
    z = (z * y) / ONE;
    n += z * 0x000000000000017c; # add y^18 * (20! / 18!)
    z = (z * y) / ONE;
    n += z * 0x0000000000000014; # add y^19 * (20! / 19!)
    z = (z * y) / ONE;
    n += z * 0x0000000000000001; # add y^20 * (20! / 20!)
    n = n / 0x21c3677c82b40000 + y + ONE; # divide by 20! and then add y^1 / 1! + y^0 / 0!

    if ((x & (ONE >> 3)) != 0):
        n = (n * 0x1c3d6a24ed82218787d624d3e5eba95f9) / 0x18ebef9eac820ae8682b9793ac6d1e776; # multiply by e^(2^-3)
    if ((x & (ONE >> 2)) != 0):
        n = (n * 0x18ebef9eac820ae8682b9793ac6d1e778) / 0x1368b2fc6f9609fe7aceb46aa619baed4; # multiply by e^(2^-2)
    if ((x & (ONE >> 1)) != 0):
        n = (n * 0x1368b2fc6f9609fe7aceb46aa619baed5) / 0x0bc5ab1b16779be3575bd8f0520a9f21f; # multiply by e^(2^-1)
    if ((x & (ONE << 0)) != 0):
        n = (n * 0x0bc5ab1b16779be3575bd8f0520a9f21e) / 0x0454aaa8efe072e7f6ddbab84b40a55c9; # multiply by e^(2^+0)
    if ((x & (ONE << 1)) != 0):
        n = (n * 0x0454aaa8efe072e7f6ddbab84b40a55c5) / 0x00960aadc109e7a3bf4578099615711ea; # multiply by e^(2^+1)
    if ((x & (ONE << 2)) != 0):
        n = (n * 0x00960aadc109e7a3bf4578099615711d7) / 0x0002bf84208204f5977f9a8cf01fdce3d; # multiply by e^(2^+2)
    if ((x & (ONE << 3)) != 0):
        n = (n * 0x0002bf84208204f5977f9a8cf01fdc307) / 0x0000003c6ab775dd0b95b4cbee7e65d11; # multiply by e^(2^+3)

    unchecked.end()

    return Fraction256({ 'n': n, 'd': ONE });

'''
    * @dev returns a fraction with truncated components
    * note that since the input value is truncated, the use of the method incurs precision loss
'''
def truncatedFraction(fraction: Fraction256, max: int) -> (Fraction256):
    scale = Math.ceilDiv(Math.max(fraction.n, fraction.d), max);
    truncated = Fraction256({ 'n': fraction.n / scale, 'd': fraction.d / scale });
    if (truncated.d == 0):
        revert("InvalidFraction");

    return truncated;

'''
    * @dev returns the weighted average of two fractions
'''
def weightedAverage(
    fraction1: Fraction256,
    fraction2: Fraction256,
    weight1,
    weight2
) -> (Fraction256):
    return Fraction256({
        'n': fraction1.n * fraction2.d * weight1 + fraction1.d * fraction2.n * weight2,
        'd': fraction1.d * fraction2.d * (weight1 + weight2)
    });

'''
    * @dev returns whether or not the deviation of an offset sample from a base sample is within a permitted range
    * for example, if the maximum permitted deviation is 5%, then evaluate `95% * base <= offset <= 105% * base`
'''
def isInRange(
    baseSample: Fraction256,
    offsetSample: Fraction256,
    maxDeviationPPM
) -> (bool):
    min = mul512(baseSample.n, offsetSample.d * (PPM_RESOLUTION - maxDeviationPPM));
    mid = mul512(baseSample.d, offsetSample.n * PPM_RESOLUTION);
    max = mul512(baseSample.n, offsetSample.d * (PPM_RESOLUTION + maxDeviationPPM));
    return lte512(min, mid) and lte512(mid, max);

'''
    * @dev returns an `Sint256` positive representation of an unsigned integer
'''
def toPos256(n) -> (Sint256):
    return Sint256({ 'value': n, 'isNeg': False });

'''
    * @dev returns an `Sint256` negative representation of an unsigned integer
'''
def toNeg256(n) -> (Sint256):
    return Sint256({ 'value': n, 'isNeg': True });

'''
    * @dev returns the largest integer smaller than or equal to `x * y / z`
'''
def mulDivF(
    x,
    y,
    z
) -> (uint):
    x, y, z = uint256(x), uint256(y), uint256(z)

    xy = mul512(x, y);

    # if `x * y < 2 ^ 256`
    if (xy.hi == 0):
        return xy.lo / z;

    # assert `x * y / z < 2 ^ 256`
    if (xy.hi >= z):
        revert("Overflow");

    m = _mulMod(x, y, z); # `m = x * y % z`
    n = _sub512(xy, m); # `n = x * y - m` hence `n / z = floor(x * y / z)`

    # if `n < 2 ^ 256`
    if (n.hi == 0):
        return n.lo / z;

    p = _unsafeSub(0, z) & z; # `p` is the largest power of 2 which `z` is divisible by
    q = _div512(n, p); # `n` is divisible by `p` because `n` is divisible by `z` and `z` is divisible by `p`
    r = _inv256(z / p); # `z / p = 1 mod 2` hence `inverse(z / p) = 1 mod 2 ^ 256`
    return _unsafeMul(q, r); # `q * r = (n / p) * inverse(z / p) = n / z`

'''
    * @dev returns the smallest integer larger than or equal to `x * y / z`
'''
def mulDivC(
    x,
    y,
    z
) -> (uint):
    x, y, z = uint256(x), uint256(y), uint256(z)

    w = mulDivF(x, y, z);
    if (_mulMod(x, y, z) > 0):
        if (w >= uint256.max):
            revert("Overflow");

        return w + 1;
    return w;

'''
    * @dev returns the maximum of `n1 - n2` and 0
'''
def subMax0(n1, n2) -> (uint):
    return uint256(n1 - n2 if n1 > n2 else 0);

'''
    * @dev returns the value of `x > y`
'''
def gt512(x: Uint512, y: Uint512) -> (bool):
    return x.hi > y.hi or (x.hi == y.hi and x.lo > y.lo);

'''
    * @dev returns the value of `x < y`
'''
def lt512(x: Uint512, y: Uint512) -> (bool):
    return x.hi < y.hi or (x.hi == y.hi and x.lo < y.lo);

'''
    * @dev returns the value of `x >= y`
'''
def gte512(x: Uint512, y: Uint512) -> (bool):
    return not lt512(x, y);

'''
    * @dev returns the value of `x <= y`
'''
def lte512(x: Uint512, y: Uint512) -> (bool):
    return not gt512(x, y);

'''
    * @dev returns the value of `x * y`
'''
def mul512(x, y) -> (Uint512):
    p = _mulModMax(x, y);
    q = _unsafeMul(x, y);
    if (p >= q):
        return Uint512({ 'hi': p - q, 'lo': q });
    return Uint512({ 'hi': _unsafeSub(p, q) - 1, 'lo': q });

'''
    * @dev returns the value of `x - y`, given that `x >= y`
'''
def _sub512(x: Uint512, y) -> (Uint512):
    if (x.lo >= y):
        return Uint512({ 'hi': x.hi, 'lo': x.lo - y });
    return Uint512({ 'hi': x.hi - 1, 'lo': _unsafeSub(x.lo, y) });

'''
    * @dev returns the value of `x / pow2n`, given that `x` is divisible by `pow2n`
'''
def _div512(x: Uint512, pow2n) -> (uint):
    pow2nInv = _unsafeAdd(_unsafeSub(0, pow2n) / pow2n, 1); # `1 << (256 - n)`
    return _unsafeMul(x.hi, pow2nInv) | (x.lo / pow2n); # `(x.hi << (256 - n)) | (x.lo >> n)`

'''
    * @dev returns the inverse of `d` modulo `2 ^ 256`, given that `d` is congruent to `1` modulo `2`
'''
def _inv256(d) -> (uint):
    # approximate the root of `f(x) = 1 / x - d` using the newton–raphson convergence method
    x = 1;
    for _ in range(8):
        x = _unsafeMul(x, _unsafeSub(2, _unsafeMul(x, d))); # `x = x * (2 - x * d) mod 2 ^ 256`
    return x;

'''
    * @dev returns `(x + y) % 2 ^ 256`
'''
def _unsafeAdd(x, y) -> (uint):
    unchecked.begin()
    z = uint256(x) + uint256(y);
    unchecked.end()
    return z

'''
    * @dev returns `(x - y) % 2 ^ 256`
'''
def _unsafeSub(x, y) -> (uint):
    unchecked.begin()
    z = uint256(x) - uint256(y);
    unchecked.end()
    return z

'''
    * @dev returns `(x * y) % 2 ^ 256`
'''
def _unsafeMul(x, y) -> (uint):
    unchecked.begin()
    z = uint256(x) * uint256(y);
    unchecked.end()
    return z

'''
    * @dev returns `x * y % (2 ^ 256 - 1)`
'''
def _mulModMax(x, y) -> (uint):
    return mulmod(x, y, uint256.max);

'''
    * @dev returns `x * y % z`
'''
def _mulMod(
    x,
    y,
    z
) -> (uint):
    return mulmod(x, y, z);

from bancor_research.bancor_emulator import config

if config.full_precision_mode:
    def exp2(f: Fraction256) -> (Fraction256):
        return Fraction256({ 'n': uint256(2) ** (f.n / f.d), 'd': 1 })
    def truncatedFraction(fraction: Fraction256, max: int) -> (Fraction256):
        return fraction
    def mulDivF(x, y, z) -> (uint):
        return uint256(x) * y / z
    def mulDivC(x, y, z) -> (uint):
        return uint256(x) * y / z
    def gt512(x: Uint512, y: Uint512) -> (bool):
        return x.lo > y.lo
    def lt512(x: Uint512, y: Uint512) -> (bool):
        return x.lo < y.lo
    def mul512(x, y) -> (Uint512):
        return Uint512({ 'hi': 0, 'lo': uint256(x) * y })

library(vars(), MathEx)


# coding=utf-8
# --------------------------------------------------------------------------------------------------------------------
# Licensed under the MIT LICENSE. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------------------------------
"""Agent-based actions. (e.g. trade, deposit, withdraw, etc...)"""
from bancor_research.bancor_simulator.v3.spec.utils import *


def deposit_bnt(state: State, tkn_name: str, tkn_amt: Decimal, user_name: str) -> State:
    """
    Specific case of `deposit` for bnt token.
    """
    bnbnt_amt = compute_bnbnt_amt(state, tkn_amt)
    protocol_wallet_balance = get_protocol_wallet_balance(state, "bnt")
    if bnbnt_amt > protocol_wallet_balance:
        ptkn_amt = bnbnt_amt - protocol_wallet_balance
        rtkn_amt = compute_rtkn_amt(state, "bnt", ptkn_amt)
        state.increase_staked_balance("bnt", rtkn_amt)
        state.increase_pooltoken_balance("bnt", ptkn_amt)
        state.increase_protocol_wallet_balance("bnt", ptkn_amt)
    state.decrease_user_balance(user_name, tkn_name, tkn_amt)
    state.decrease_protocol_wallet_balance("bnt", bnbnt_amt)
    state.increase_user_balance(user_name, f"bn{tkn_name}", bnbnt_amt)
    state.increase_user_balance(user_name, "vbnt", bnbnt_amt)
    return state


def deposit_tkn(state: State, tkn_name: str, tkn_amt: Decimal, user_name: str) -> State:
    """
    Specific case of `deposit` for non-bnt token.
    """
    bntkn_amt = compute_bntkn_amt(state, tkn_name, tkn_amt)
    state.decrease_user_balance(user_name, tkn_name, tkn_amt)
    state.increase_user_balance(user_name, f"bn{tkn_name}", bntkn_amt)
    state.increase_pooltoken_balance(tkn_name, bntkn_amt)
    state.increase_master_vault_balance(tkn_name, tkn_amt)
    state.increase_staked_balance(tkn_name, tkn_amt)

    case, bnt_increase, tkn_increase = compute_pool_depth_adjustment(state, tkn_name)

    bnbnt_amt = compute_bnbnt_amt(state, bnt_increase)
    state.increase_bnt_trading_liquidity(tkn_name, bnt_increase)
    state.increase_tkn_trading_liquidity(tkn_name, tkn_increase)
    state.increase_bnt_funding_amt(tkn_name, bnt_increase)
    state.increase_protocol_wallet_balance("bnt", bnbnt_amt)
    state.increase_pooltoken_balance("bnt", bnbnt_amt)
    state.increase_master_vault_balance("bnt", bnt_increase)
    state.increase_staked_balance("bnt", bnt_increase)

    state.update_spot_rate(tkn_name)

    if get_is_trading_enabled(state, tkn_name):
        state = handle_ema(state, tkn_name)

    if check_pool_shutdown(state, tkn_name) and bnt_increase > 0:
        state = shutdown_pool(state, tkn_name)

    return state


def process_trade(
        state: State,
        tkn_amt: Decimal,
        source_token: str,
        target_token: str,
        user_name: str,
        timestamp: int = 0,
) -> State:
    """
    Main logic for trade actions.
    """

    is_trading_enabled_target = get_is_trading_enabled(state, target_token)
    is_trading_enabled_source = get_is_trading_enabled(state, source_token)

    if source_token == "bnt" and is_trading_enabled_target:

        (
            tkn_name,
            bnt_trading_liquidity,
            tkn_trading_liquidity,
            trading_fee,
            network_fee,
        ) = get_trade_inputs(state, target_token)

        state, target_sent_to_user = trade_bnt_for_tkn(
            state,
            bnt_trading_liquidity,
            tkn_trading_liquidity,
            trading_fee,
            network_fee,
            tkn_amt,
            tkn_name,
        )

    elif target_token == "bnt" and is_trading_enabled_source:

        (
            tkn_name,
            bnt_trading_liquidity,
            tkn_trading_liquidity,
            trading_fee,
            network_fee,
        ) = get_trade_inputs(state, source_token)

        state, target_sent_to_user = trade_tkn_for_bnt(
            state,
            bnt_trading_liquidity,
            tkn_trading_liquidity,
            trading_fee,
            network_fee,
            tkn_amt,
            tkn_name,
        )

    elif (
            source_token != "bnt"
            and target_token != "bnt"
            and is_trading_enabled_source
            and is_trading_enabled_target
    ):

        (
            tkn_name,
            bnt_trading_liquidity,
            tkn_trading_liquidity,
            trading_fee,
            network_fee,
        ) = get_trade_inputs(state, source_token)

        state, intermediate_bnt = trade_tkn_for_bnt(
            state,
            bnt_trading_liquidity,
            tkn_trading_liquidity,
            trading_fee,
            network_fee,
            tkn_amt,
            tkn_name,
        )

        (
            tkn_name,
            bnt_trading_liquidity,
            tkn_trading_liquidity,
            trading_fee,
            network_fee,
        ) = get_trade_inputs(state, target_token)

        state, target_sent_to_user = trade_bnt_for_tkn(
            state,
            bnt_trading_liquidity,
            tkn_trading_liquidity,
            trading_fee,
            network_fee,
            intermediate_bnt,
            tkn_name,
        )

    else:
        # Trading is disabled
        tkn_amt = Decimal("0")
        target_sent_to_user = Decimal("0")

    state.decrease_user_balance(user_name, source_token, tkn_amt)
    state.increase_user_balance(user_name, target_token, target_sent_to_user)
    return state


def trade_bnt_for_tkn(
        state: State,
        bnt_trading_liquidity: Decimal,
        tkn_trading_liquidity: Decimal,
        trading_fee: Decimal,
        network_fee: Decimal,
        bnt_amt: Decimal,
        tkn_name: str,
        direction: str = "bnt",
) -> Tuple[State, Decimal]:
    """
    Specific case of `trade` when bnt is the source token.
    """
    state = handle_ema(state, tkn_name)
    updated_bnt_liquidity = compute_changed_bnt_trading_liquidity(
        bnt_trading_liquidity,
        tkn_trading_liquidity,
        trading_fee,
        network_fee,
        bnt_amt,
        direction,
    )

    updated_tkn_liquidity = compute_changed_tkn_trading_liquidity(
        bnt_trading_liquidity, tkn_trading_liquidity, trading_fee, bnt_amt, direction
    )
    tkn_sent_to_user = compute_target_amt(
        bnt_trading_liquidity, tkn_trading_liquidity, trading_fee, bnt_amt, direction
    )
    trade_fee = swap_fee_collection(
        bnt_trading_liquidity,
        tkn_trading_liquidity,
        trading_fee,
        network_fee,
        bnt_amt,
        direction,
    )

    if tkn_name in state.rolling_trade_fees:
        state.rolling_trade_fees[tkn_name].append(trade_fee)
    else:
        state.rolling_trade_fees[tkn_name] = []
        state.rolling_trade_fees[tkn_name].append(trade_fee)

    bnt_collected_by_vortex = vortex_collection(
        bnt_trading_liquidity,
        tkn_trading_liquidity,
        trading_fee,
        network_fee,
        bnt_amt,
        direction,
    )

    state.increase_master_vault_balance("bnt", bnt_amt)
    state.decrease_master_vault_balance(tkn_name, tkn_sent_to_user)
    state.set_bnt_trading_liquidity(tkn_name, updated_bnt_liquidity)
    state.set_tkn_trading_liquidity(tkn_name, updated_tkn_liquidity)
    state.increase_staked_balance(tkn_name, trade_fee)
    state.increase_vortex_balance("bnt", bnt_collected_by_vortex)
    state.update_spot_rate(tkn_name)
    return state, tkn_sent_to_user


def trade_tkn_for_bnt(
        state: State,
        bnt_trading_liquidity: Decimal,
        tkn_trading_liquidity: Decimal,
        trading_fee: Decimal,
        network_fee: Decimal,
        tkn_amt: Decimal,
        tkn_name: str,
        direction="tkn",
):
    """
    Specific case of `trade` when bnt is the target token.
    """

    state = handle_ema(state, tkn_name)
    updated_bnt_liquidity = compute_changed_bnt_trading_liquidity(
        bnt_trading_liquidity,
        tkn_trading_liquidity,
        trading_fee,
        network_fee,
        tkn_amt,
        direction,
    )
    updated_tkn_liquidity = compute_changed_tkn_trading_liquidity(
        bnt_trading_liquidity, tkn_trading_liquidity, trading_fee, tkn_amt, direction
    )
    bnt_sent_to_user = compute_target_amt(
        bnt_trading_liquidity, tkn_trading_liquidity, trading_fee, tkn_amt, direction
    )
    trade_fee = swap_fee_collection(
        bnt_trading_liquidity,
        tkn_trading_liquidity,
        trading_fee,
        network_fee,
        tkn_amt,
        direction,
    )

    if tkn_name in state.rolling_trade_fees:
        state.rolling_trade_fees[tkn_name].append(trade_fee)
    else:
        state.rolling_trade_fees[tkn_name] = []
        state.rolling_trade_fees[tkn_name].append(trade_fee)

    bnt_collected_by_vortex = vortex_collection(
        bnt_trading_liquidity,
        tkn_trading_liquidity,
        trading_fee,
        network_fee,
        tkn_amt,
        direction,
    )

    state.decrease_master_vault_balance("bnt", bnt_sent_to_user)
    state.increase_master_vault_balance(tkn_name, tkn_amt)
    state.set_bnt_trading_liquidity(tkn_name, updated_bnt_liquidity)
    state.set_tkn_trading_liquidity(tkn_name, updated_tkn_liquidity)
    state.increase_staked_balance("bnt", trade_fee)
    state.increase_bnt_funding_amt(tkn_name, trade_fee)
    state.increase_vortex_balance("bnt", bnt_collected_by_vortex)
    state.update_spot_rate(tkn_name)
    return state, bnt_sent_to_user


def process_withdrawal(
        state: State,
        user_name: str,
        id_number: int,
        timestamp: int = 0,
):
    """
    Main withdrawal logic based on the withdraw algorithm of the BIP15 spec.
    """

    (
        id_number,
        cooldown_timestamp,
        tkn_name,
        pool_token_amt,
        withdraw_value,
        user_name,
    ) = unpack_withdrawal_cooldown(state, user_name, id_number)

    withdrawal_fee = state.withdrawal_fee
    cooldown_time = state.tokens[tkn_name].cooldown_time
    cool_down_complete = timestamp - cooldown_timestamp >= cooldown_time

    if tkn_name != "bnt":
        bnbnt_rate = state.bnbnt_rate
        is_price_stable = get_is_price_stable(state, tkn_name)

        if cool_down_complete and is_price_stable:

            state.set_pending_withdrawals_status(user_name, id_number, True)

            bnt_trading_liquidity = get_bnt_trading_liquidity(state, tkn_name)
            tkn_trading_liquidity = get_tkn_trading_liquidity(state, tkn_name)
            tkn_excess = get_tkn_excess(state, tkn_name)
            staked_tkn = get_staked_balance(state, tkn_name)
            external_protection_tkn_balance = get_external_protection_vault_balance(
                state, tkn_name
            )
            trading_fee = get_trading_fee(state, tkn_name)

            solver = WithdrawalAlgorithm(
                bnt_trading_liquidity=bnt_trading_liquidity,
                # changed due to contract implementation
                # was previously `avg_tkn_trading_liquidity`
                tkn_trading_liquidity=tkn_trading_liquidity,
                tkn_excess=tkn_excess,
                staked_tkn=staked_tkn,
                external_protection_tkn_balance=external_protection_tkn_balance,
                trading_fee=trading_fee,
                withdrawal_fee=withdrawal_fee,
                withdraw_value=withdraw_value,
            )
            (
                updated_bnt_liquidity,
                bnt_renounced,
                updated_tkn_liquidity,
                tkn_sent_to_user,
                bnt_sent_to_user,
                external_protection_compensation,
            ) = solver.process_withdrawal()

            if bnt_renounced == 0:
                bnt_delta = updated_bnt_liquidity - bnt_trading_liquidity
                state.increase_master_vault_balance("bnt", bnt_delta)

            state.set_bnt_trading_liquidity(tkn_name, updated_bnt_liquidity)
            state.decrease_master_vault_balance("bnt", bnt_renounced)

            bnt_renounced = min(bnt_renounced, get_bnt_funding_amt(state, tkn_name))
            state.decrease_bnt_funding_amt(tkn_name, bnt_renounced)
            state.decrease_staked_balance("bnt", bnt_renounced)

            bnbnt_renounced = bnbnt_rate * bnt_renounced

            state.decrease_pooltoken_balance("bnt", bnbnt_renounced)
            state.decrease_protocol_wallet_balance("bnt", bnbnt_renounced)
            state.set_tkn_trading_liquidity(tkn_name, updated_tkn_liquidity)
            state.decrease_pooltoken_balance(tkn_name, pool_token_amt)
            state.decrease_staked_balance(tkn_name, withdraw_value)
            state.decrease_master_vault_balance(tkn_name, tkn_sent_to_user)
            state.increase_user_balance(
                user_name,
                tkn_name,
                tkn_sent_to_user + external_protection_compensation,
            )
            state.decrease_external_protection_balance(
                tkn_name, external_protection_compensation
            )
            state.increase_user_balance(user_name, "bnt", bnt_sent_to_user)

            state.update_spot_rate(tkn_name)

            if get_is_trading_enabled(state, tkn_name):
                state = handle_ema(state, tkn_name)

            if check_pool_shutdown(state, tkn_name):
                state = shutdown_pool(state, tkn_name)

    else:

        sufficient_vbnt = get_user_balance(state, user_name, "vbnt") >= pool_token_amt

        if cool_down_complete and sufficient_vbnt:
            bnt_amt = withdraw_value * (1 - withdrawal_fee)
            state.set_pending_withdrawals_status(user_name, id_number, True)
            state.decrease_user_balance(user_name, "vbnt", pool_token_amt)
            state.increase_user_balance(user_name, "bnt", bnt_amt)
            state.increase_protocol_wallet_balance("bnt", pool_token_amt)

    return state


def begin_withdrawal_cooldown_by_rtkn(state, rtkn_amt, tkn_name, user_name):
    """
    Begin the withdrawal cooldown operation by specifying the output amount of reserve tokens.
    """
    ptkn_amt = compute_ptkn_amt(state, tkn_name, rtkn_amt)
    return _begin_withdrawal_cooldown(state, tkn_name, user_name, rtkn_amt, ptkn_amt)


def begin_withdrawal_cooldown_by_ptkn(state, ptkn_amt, tkn_name, user_name):
    """
    Begin the withdrawal cooldown operation by specifying the intput amount of pool tokens.
    """
    rtkn_amt = compute_rtkn_amt(state, tkn_name, ptkn_amt)
    return _begin_withdrawal_cooldown(state, tkn_name, user_name, rtkn_amt, ptkn_amt)


def _begin_withdrawal_cooldown(state, tkn_name, user_name, rtkn_amt, ptkn_amt):
    """
    After a fixed time duration, these items can be retrieved and passed to the withdrawal algorithm.
    """
    id_number = get_withdrawal_id(state)
    state.users[user_name].pending_withdrawals[id_number] = Cooldown(
        id=id_number,
        created_at=state.timestamp,
        user_name=user_name,
        tkn_name=tkn_name,
        tkn=Token(balance=rtkn_amt),
        pooltoken=Token(balance=ptkn_amt),
        is_complete=False,
    )

    state.decrease_user_balance(user_name, f"bn{tkn_name}", ptkn_amt)
    return id_number


def unpack_withdrawal_cooldown(
        state: State, user_name: str, id_number: int
) -> Tuple[int, int, str, Decimal, Decimal, str]:
    """
    Introduced to make the withdrawals easier to handle.
    """
    if id_number in state.users[user_name].pending_withdrawals:
        cool_down_state = state.users[user_name].pending_withdrawals[id_number]
        if not cool_down_state.is_complete:
            cooldown_timestamp = cool_down_state.created_at
            tkn_name = cool_down_state.tkn_name
            pool_token_amt = cool_down_state.pooltoken.balance
            withdrawal_value = cool_down_state.tkn.balance
            return (
                id_number,
                cooldown_timestamp,
                tkn_name,
                pool_token_amt,
                withdrawal_value,
                user_name,
            )
    raise ValueError(
        f"id_number={id_number} not foound in user_name={user_name} wallet pending_withdrawals"
    )


@dataclass
class WithdrawalAlgorithm:
    bnt_trading_liquidity: Decimal = Decimal("0")
    tkn_trading_liquidity: Decimal = Decimal("0")
    tkn_excess: Decimal = Decimal("0")
    staked_tkn: Decimal = Decimal("0")
    trading_fee: Decimal = Decimal("0")
    withdrawal_fee: Decimal = Decimal("0")
    withdraw_value: Decimal = Decimal("0")
    bnt_sent_to_user: Decimal = Decimal("0")
    external_protection_tkn_balance: Decimal = Decimal("0")
    tkn_sent_to_user: Decimal = Decimal("0")
    is_trading_enabled: bool = True

    @property
    def hmax_surplus(self):
        """
        Returns the maximum number of withdrawn TKN that can currently be supported under the arbitrage method.
        """
        b = self.tkn_trading_liquidity
        c = self.tkn_excess
        e = self.staked_tkn
        m = self.trading_fee
        n = self.withdrawal_fee
        hmax = (
                b
                * e
                * (e * n + m * (b + c - e))
                / ((1 - m) * (b + c - e) * (b + c - e * (1 - n)))
        )
        return hmax

    @property
    def hmax_deficit(self):
        """
        Returns the maximum number of withdrawn TKN that can currently be supported under the arbitrage method.
        """
        b = self.tkn_trading_liquidity
        c = self.tkn_excess
        e = self.staked_tkn
        m = self.trading_fee
        n = self.withdrawal_fee
        hmax = (
                b
                * e
                * (e * n - m * (b + c - e * (1 - n)))
                / ((1 - m) * (b + c - e) * (b + c - e * (1 - n)))
        )
        return 0  # hmax (retroactively, due to change in contract implementation)

    @property
    def hlim(self):
        """
        Returns the maximum number of withdrawn TKN that can currently be supported under the arbitrage method.
        """
        b = self.tkn_trading_liquidity
        c = self.tkn_excess
        e = self.staked_tkn
        hlim = c * e / (b + c)
        return hlim

    def arbitrage_withdrawal_surplus(self):
        """
        Calculates the withdrawal output variables when the v3 can arbitrage the pool.
        """
        a = self.bnt_trading_liquidity
        b = self.tkn_trading_liquidity
        c = self.tkn_excess
        e = self.staked_tkn
        m = self.trading_fee
        n = self.withdrawal_fee
        x = self.withdraw_value
        updated_bnt_liquidity = (
                a
                * (b * e - m * (b * e + x * (b + c - e * (1 - n))))
                / ((1 - m) * (b * e + x * (b + c - e * (1 - n))))
        )
        bnt_renounced = Decimal("0")
        updated_tkn_liquidity = (b * e + x * (b + c + e * (n - 1))) / e
        tkn_sent_to_user = x * (1 - n)
        bnt_sent_to_user = Decimal("0")
        return (
            updated_bnt_liquidity,
            bnt_renounced,
            updated_tkn_liquidity,
            tkn_sent_to_user,
            bnt_sent_to_user,
        )

    def default_withdrawal_surplus_covered(
            self,
    ):
        """
        Calculates the withdrawal output variables when arbitrage is forbidden, and the excess TKN is sufficient to cover the entire withdrawal amt.
        """
        x = self.withdraw_value
        n = self.withdrawal_fee
        updated_bnt_liquidity = self.bnt_trading_liquidity
        bnt_renounced = Decimal("0")
        updated_tkn_liquidity = self.tkn_trading_liquidity
        tkn_sent_to_user = x * (1 - n)
        bnt_sent_to_user = Decimal("0")
        return (
            updated_bnt_liquidity,
            bnt_renounced,
            updated_tkn_liquidity,
            tkn_sent_to_user,
            bnt_sent_to_user,
        )

    def default_withdrawal_surplus_exposed(self):
        """
        Calculates the withdrawal output variables when arbitrage is forbidden, and the TKN trading liquidity must be used to cover the withdrawal amt.
        """
        a = self.bnt_trading_liquidity
        b = self.tkn_trading_liquidity
        c = self.tkn_excess
        n = self.withdrawal_fee
        x = self.withdraw_value
        updated_bnt_liquidity = a * (b + c - x * (1 - n)) / b
        bnt_renounced = a * (x * (1 - n) - c) / b
        updated_tkn_liquidity = b + c - x * (1 - n)
        tkn_sent_to_user = x * (1 - n)
        bnt_sent_to_user = Decimal("0")
        return (
            updated_bnt_liquidity,
            bnt_renounced,
            updated_tkn_liquidity,
            tkn_sent_to_user,
            bnt_sent_to_user,
        )

    def arbitrage_withdrawal_deficit(self):
        """
        Calculates the withdrawal output variables when the v3 can arbitrage the pool.
        """
        a = self.bnt_trading_liquidity
        b = self.tkn_trading_liquidity
        c = self.tkn_excess
        e = self.staked_tkn
        m = self.trading_fee
        n = self.withdrawal_fee
        x = self.withdraw_value
        updated_bnt_liquidity = (
                a * b * e / (b * e + x * (1 - m) * (b + c - e * (1 - n)))
        )
        bnt_renounced = Decimal("0")
        updated_tkn_liquidity = (b * e + x * (b + c - e * (1 - n))) / e
        tkn_sent_to_user = x * (1 - n)
        bnt_sent_to_user = Decimal("0")
        return (
            updated_bnt_liquidity,
            bnt_renounced,
            updated_tkn_liquidity,
            tkn_sent_to_user,
            bnt_sent_to_user,
        )

    def default_withdrawal_deficit_covered(self):
        """
        Calculates the withdrawal output variables when arbitrage is forbidden, and the excess TKN is sufficient to cover the entire withdrawal amt.
        """
        a = self.bnt_trading_liquidity
        b = self.tkn_trading_liquidity
        c = self.tkn_excess
        e = self.staked_tkn
        n = self.withdrawal_fee
        x = self.withdraw_value
        updated_bnt_liquidity = a
        bnt_renounced = Decimal("0")
        updated_tkn_liquidity = self.tkn_trading_liquidity
        tkn_sent_to_user = x * (1 - n) * (b + c) / e
        bnt_sent_to_user = a * x * (1 - n) * (e - b - c) / (b * e)
        return (
            updated_bnt_liquidity,
            bnt_renounced,
            updated_tkn_liquidity,
            tkn_sent_to_user,
            bnt_sent_to_user,
        )

    def default_withdrawal_deficit_exposed(self):
        """
        Calculates the withdrawal output variables when arbitrage is forbidden, and the excess TKN is sufficient to cover the entire withdrawal amt.
        """
        a = self.bnt_trading_liquidity
        b = self.tkn_trading_liquidity
        c = self.tkn_excess
        e = self.staked_tkn
        n = self.withdrawal_fee
        x = self.withdraw_value
        updated_bnt_liquidity = a * (b + c) * (e - x * (1 - n)) / (b * e)
        bnt_renounced = a * (b * e - (b + c) * (e - x * (1 - n))) / (b * e)
        updated_tkn_liquidity = (b + c) * (e - x * (1 - n)) / e
        tkn_sent_to_user = x * (1 - n) * (b + c) / e
        bnt_sent_to_user = a * x * (1 - n) * (e - b - c) / (b * e)
        return (
            updated_bnt_liquidity,
            bnt_renounced,
            updated_tkn_liquidity,
            tkn_sent_to_user,
            bnt_sent_to_user,
        )

    def external_protection(
            self,
            bnt_trading_liquidity,
            tkn_trading_liquidity,
            bnt_sent_to_user,
            tkn_sent_to_user,
    ):
        """
        This replaces any BNT that would have been received by the user with TKN.
        """
        a = bnt_trading_liquidity
        b = tkn_trading_liquidity
        n = self.withdrawal_fee
        T = bnt_sent_to_user
        w = self.external_protection_tkn_balance
        x = self.withdraw_value
        S = tkn_sent_to_user

        if not self.is_trading_enabled:
            bnt_sent_to_user = Decimal("0")
            external_protection_compensation = min(w, x * (1 - n) - S)
        elif T and w:
            if T * b > w * a:
                bnt_sent_to_user = (T * b - w * a) / b
                external_protection_compensation = w
            else:
                bnt_sent_to_user = Decimal("0")
                external_protection_compensation = T * b / a
        else:
            bnt_sent_to_user = T
            external_protection_compensation = Decimal("0")
        return bnt_sent_to_user, external_protection_compensation

    def illiquid_withdrawal(self):
        """
        Calculates the withdrawal output variables when trading is disabled.
        """
        c = self.tkn_excess
        e = self.staked_tkn
        n = self.withdrawal_fee
        x = self.withdraw_value
        updated_bnt_liquidity = self.bnt_trading_liquidity
        bnt_renounced = Decimal("0")
        updated_tkn_liquidity = self.tkn_trading_liquidity
        if c / (1 - n) > e:
            tkn_sent_to_user = x * (1 - n)
        else:
            tkn_sent_to_user = c * x * (1 - n) / e
        bnt_sent_to_user = Decimal("0")
        return (
            updated_bnt_liquidity,
            bnt_renounced,
            updated_tkn_liquidity,
            tkn_sent_to_user,
            bnt_sent_to_user,
        )

    def process_withdrawal(
            self,
            updated_bnt_liquidity: Decimal = Decimal("0"),
            bnt_renounced: Decimal = Decimal("0"),
            updated_tkn_liquidity: Decimal = Decimal("0"),
            tkn_sent_to_user: Decimal = Decimal("0"),
    ):
        """
        Evaluates the conditions, and calls the appropriate functions
        """

        b = self.tkn_trading_liquidity
        c = self.tkn_excess
        e = self.staked_tkn
        n = self.withdrawal_fee
        x = self.withdraw_value

        bnt_sent_to_user = Decimal("0")
        external_protection_compensation = Decimal("0")

        if not self.is_trading_enabled:
            (
                updated_bnt_liquidity,
                bnt_renounced,
                updated_tkn_liquidity,
                tkn_sent_to_user,
                bnt_sent_to_user,
            ) = self.illiquid_withdrawal()
        elif b + c > e * (1 - n):
            if b + c > e and x < self.hlim and x < self.hmax_surplus:
                (
                    updated_bnt_liquidity,
                    bnt_renounced,
                    updated_tkn_liquidity,
                    tkn_sent_to_user,
                    bnt_sent_to_user,
                ) = self.arbitrage_withdrawal_surplus()
            elif x * (1 - n) <= c:
                (
                    updated_bnt_liquidity,
                    bnt_renounced,
                    updated_tkn_liquidity,
                    tkn_sent_to_user,
                    bnt_sent_to_user,
                ) = self.default_withdrawal_surplus_covered()
            else:
                (
                    updated_bnt_liquidity,
                    bnt_renounced,
                    updated_tkn_liquidity,
                    tkn_sent_to_user,
                    bnt_sent_to_user,
                ) = self.default_withdrawal_surplus_exposed()
        elif b + c <= e * (1 - n):
            if x < self.hlim and x < self.hmax_deficit:
                (
                    updated_bnt_liquidity,
                    bnt_renounced,
                    updated_tkn_liquidity,
                    tkn_sent_to_user,
                    bnt_sent_to_user,
                ) = self.arbitrage_withdrawal_deficit()
            elif x * (1 - n) * (b + c) <= c * e:
                (
                    updated_bnt_liquidity,
                    bnt_renounced,
                    updated_tkn_liquidity,
                    tkn_sent_to_user,
                    bnt_sent_to_user,
                ) = self.default_withdrawal_deficit_covered()
            else:
                (
                    updated_bnt_liquidity,
                    bnt_renounced,
                    updated_tkn_liquidity,
                    tkn_sent_to_user,
                    bnt_sent_to_user,
                ) = self.default_withdrawal_deficit_exposed()
            (
                bnt_sent_to_user,
                external_protection_compensation,
            ) = self.external_protection(
                updated_bnt_liquidity,
                updated_tkn_liquidity,
                bnt_sent_to_user,
                tkn_sent_to_user,
            )

        return (
            updated_bnt_liquidity,
            bnt_renounced,
            updated_tkn_liquidity,
            tkn_sent_to_user,
            bnt_sent_to_user,
            external_protection_compensation,
        )


def get_burn_reward(vortex_ledger_balance: Decimal) -> Tuple[Decimal, Decimal]:
    """
    Returns the burn_reward and the updated_bnt_amount.
    """
    burn_reward = min(Decimal("100"), vortex_ledger_balance / Decimal("10"))
    bnt_amount = vortex_ledger_balance - burn_reward
    return burn_reward, bnt_amount


def get_target_amount(a, b, d, x, direction):
    """
    Returns the quantity of the target asset purchased to the user.
    """
    if direction == "tkn":
        return a * x * (1 - d) / (b + x)
    elif direction == "bnt":
        return b * x * (1 - d) / (a + x)


def vortex_burner(state: State, user_name: str):
    """
    The bnt balance of the vortex is swapped for vbnt, vbnt is destroyed, and the user is rewarded.
    """
    vortex_ledger_balance = get_vortex_balance(state, "bnt")
    tkn_name = "vbnt"
    if get_is_trading_enabled(state, tkn_name) and vortex_ledger_balance > 0:
        burn_reward, bnt_amount = get_burn_reward(vortex_ledger_balance)
        handle_ema(tkn_name)
        bnt_trading_liquidity = get_bnt_trading_liquidity(state, tkn_name)
        tkn_trading_liquidity = get_tkn_trading_liquidity(state, tkn_name)
        trading_fee = get_trading_fee(state, tkn_name)
        network_fee = Decimal("0")
        direction = "bnt"
        updated_bnt_liquidity = compute_changed_bnt_trading_liquidity(
            bnt_trading_liquidity,
            tkn_trading_liquidity,
            trading_fee,
            network_fee,
            bnt_amount,
            direction,
        )
        updated_tkn_liquidity = compute_changed_tkn_trading_liquidity(
            bnt_trading_liquidity,
            tkn_trading_liquidity,
            trading_fee,
            bnt_amount,
            direction,
        )
        vbnt_to_burner = vortex_collection(
            bnt_trading_liquidity,
            tkn_trading_liquidity,
            trading_fee,
            bnt_amount,
            direction,
        )
        trade_fee = swap_fee_collection(
            bnt_trading_liquidity,
            tkn_trading_liquidity,
            trading_fee,
            network_fee,
            bnt_amount,
            direction,
        )
        state.increase_user_balance(user_name, "bnt", burn_reward)
        state.set_vortex_balance("bnt", Decimal("0"))
        state.increase_vbnt_burned("bnt", Decimal("0"))
        state.decrease_master_vault_balance("bnt", burn_reward)
        state.decrease_master_vault_balance(tkn_name, burn_reward)
        state.set_bnt_trading_liquidity(tkn_name, updated_bnt_liquidity)
        state.set_tkn_trading_liquidity(tkn_name, updated_tkn_liquidity)
        state.increase_staked_balance(tkn_name, trade_fee)
        state.update_spot_rate(tkn_name)


def process_arbitrage_trade(
        tkn_name: str,
        tkn_token_virtual_balance: Decimal,
        bnt_virtual_balance: Decimal,
        bnt_trading_liquidity: Decimal,
        tkn_trading_liquidity: Decimal,
        trading_fee: Decimal,
        user_tkn: Decimal,
        user_bnt: Decimal,
) -> Tuple[Decimal, str, str, bool]:
    """
    Computes the appropriate arbitrage trade on the tkn_name pool.
    """
    a = bnt_trading_liquidity
    b = tkn_trading_liquidity
    m = trading_fee
    p = bnt_virtual_balance
    q = tkn_token_virtual_balance

    bnt_trade_amt = (
                            -Decimal("2") * a * q
                            + b * m * p
                            + (
                                    (Decimal("2") * a * q - b * m * p) ** Decimal("2")
                                    - Decimal("4") * a * q * (a * q - b * p)
                            )
                            ** (Decimal("1") / Decimal("2"))
                    ) / (Decimal("2") * q)

    tkn_trade_amt = (
                            -Decimal("2") * b * p
                            + a * m * q
                            + (
                                    (Decimal("2") * b * p - a * m * q) ** Decimal("2")
                                    - Decimal("4") * b * p * (b * p - a * q)
                            )
                            ** (Decimal("1") / Decimal("2"))
                    ) / (Decimal("2") * p)

    if bnt_trade_amt > 0:
        source_token = "bnt"
        target_token = tkn_name
        trade_amt = bnt_trade_amt
        user_capability = user_bnt > bnt_trade_amt
        return trade_amt, source_token, target_token, user_capability

    elif tkn_trade_amt > 0:
        source_token = tkn_name
        target_token = "bnt"
        trade_amt = tkn_trade_amt
        user_capability = user_tkn > tkn_trade_amt
        return trade_amt, source_token, target_token, user_capability


def trade_tkn_to_ema(
        bnt_trading_liquidity: Decimal,
        tkn_trading_liquidity: Decimal,
        trading_fee: Decimal,
        network_fee: Decimal,
        future_ema: Decimal,
) -> Decimal:
    """
    Outputs the tkn_amt that should be traded to force the ema and the spot price together on a given pool.
    """
    a = bnt_trading_liquidity
    b = tkn_trading_liquidity
    d = trading_fee
    e = network_fee
    f = future_ema
    tkn_amt = (
                      (a * d * (Decimal("1") - e) - Decimal("2") * f * b)
                      + (
                              a
                              * (
                                      Decimal("4") * f * b * (Decimal("1") - d * (Decimal("1") - e))
                                      + a * d ** Decimal("2") * (Decimal("1") - e) ** Decimal("2")
                              )
                      )
                      ** (Decimal("1") / Decimal("2"))
              ) / (Decimal("2") * f)
    return tkn_amt


def trade_bnt_to_ema(
        bnt_trading_liquidity,
        tkn_trading_liquidity,
        trading_fee,
        network_fee,
        future_ema,
):
    """
    Analyze the state of any pool and create a swap that drives the ema and the spot price together.
    """

    a = bnt_trading_liquidity
    b = tkn_trading_liquidity
    d = trading_fee
    e = network_fee
    f = future_ema
    x = (
                -Decimal("2") * a
                + b * d * f
                + (
                        (Decimal("2") * a - b * d * f) ** Decimal("2")
                        - Decimal("4") * a * (a - b * f)
                )
                ** (Decimal("1") / Decimal("2"))
        ) / Decimal("2")
    a_recursion = (
                          a * (a + x) + d * (Decimal("1") - e) * (a * x + x ** Decimal("2"))
                  ) / (a + d * x)
    b_recursion = b * (a + d * x) / (a + x)
    n = 0
    p = Decimal("0.001")
    while a_recursion / b_recursion < f:
        n += 1
        p += Decimal("0.0001")
        x += x * (f ** p - (a_recursion / b_recursion) ** p) / f
        a_recursion = (
                              a * (a + x) + d * (Decimal("1") - e) * (a * x + x ** Decimal("2"))
                      ) / (a + d * x)
        b_recursion = b * (a + d * x) / (a + x)
        if n > 20000:
            break
    bnt_amt = x
    return bnt_amt


def process_force_moving_average(
        self, tkn_name: str, user_tkn: Decimal, user_bnt: Decimal
) -> Tuple[Decimal, str, str, bool]:
    """
    Process the trade amount to force the ema and the spot price together.
    """
    state = self.protocol.v3.global_state
    tkn_trading_liquidity = get_tkn_trading_liquidity(state, tkn_name)
    bnt_trading_liquidity = get_bnt_trading_liquidity(state, tkn_name)
    trading_fee = get_trading_fee(state, tkn_name)
    network_fee = get_network_fee(state, tkn_name)
    ema_rate = get_ema_rate(state, tkn_name)
    spot_rate = get_spot_rate(state, tkn_name)
    future_ema = compute_ema(spot_rate, ema_rate)
    if ema_rate < spot_rate:
        source_tkn = tkn_name
        target_tkn = "bnt"
        trade_amt = trade_tkn_to_ema(
            bnt_trading_liquidity,
            tkn_trading_liquidity,
            trading_fee,
            network_fee,
            future_ema,
        )
        user_capability = user_tkn > trade_amt
    elif ema_rate > spot_rate:
        source_tkn = "bnt"
        target_tkn = tkn_name
        trade_amt = trade_bnt_to_ema(
            bnt_trading_liquidity,
            tkn_trading_liquidity,
            trading_fee,
            network_fee,
            future_ema,
        )
        user_capability = user_bnt > trade_amt
    else:
        source_tkn = tkn_name
        target_tkn = tkn_name
        trade_amt = Decimal("0")
        user_capability = False

    self.latest_tkn_name = source_tkn + "_" + target_tkn
    self.latest_amt = trade_amt
    return trade_amt, source_tkn, target_tkn, user_capability



from bancor_research.bancor_emulator.solidity import uint, uint32, uint256, address, mapping, days, revert
from bancor_research.bancor_emulator.utils import contract, parse

from bancor_research.bancor_emulator.EnumerableSet import EnumerableSet
from bancor_research.bancor_emulator.Time import Time

'''
 * @dev the data struct representing a pending withdrawal request
'''
class WithdrawalRequest:
    def __init__(self, x = None) -> None:
        self.provider = parse(address, x, 'provider'); # the liquidity provider
        self.poolToken = parse(address, x, 'poolToken'); # the locked pool token
        self.reserveToken = parse(address, x, 'reserveToken'); # the reserve token to withdraw
        self.createdAt = parse(uint32, x, 'createdAt'); # the time when the request was created (Unix timestamp)
        self.poolTokenAmount = parse(uint256, x, 'poolTokenAmount'); # the locked pool token amount
        self.reserveTokenAmount = parse(uint256, x, 'reserveTokenAmount'); # the expected reserve token amount to withdraw

'''
 * @dev the data struct representing a completed withdrawal request
'''
class CompletedWithdrawal:
    def __init__(self, x = None) -> None:
        self.poolToken = parse(address, x, 'poolToken'); # the withdraw pool token
        self.poolTokenAmount = parse(uint256, x, 'poolTokenAmount'); # the original pool token amount in the withdrawal request
        self.reserveTokenAmount = parse(uint256, x, 'reserveTokenAmount'); # the original reserve token amount at the time of the withdrawal init request

'''
 * @dev Pending Withdrawals contract
'''
class PendingWithdrawals(contract, Time):
    DEFAULT_LOCK_DURATION = uint32(7 * days);

    '''
     * @dev a "virtual" constructor that is only used to set immutable state variables
    '''
    def __init__(self,
        initNetwork,
        initBNT,
        initBNTPool
    ) -> None:
        contract.__init__(self)
        Time.__init__(self)

        self._network = initNetwork;
        self._bnt = initBNT;
        self._bntPool = initBNTPool;

    '''
     * @dev fully initializes the contract and its parents
    '''
    def initialize(self) -> None:
        self.__PendingWithdrawals_init();

    # solhint-disable func-name-mixedcase

    '''
     * @dev initializes the contract and its parents
    '''
    def __PendingWithdrawals_init(self) -> None:
        self.__PendingWithdrawals_init_unchained();

    '''
     * @dev performs contract-specific initialization
    '''
    def __PendingWithdrawals_init_unchained(self) -> None:
        # the lock duration
        self._lockDuration = uint32()

        # a mapping between accounts and their pending withdrawal requests
        self._nextWithdrawalRequestId = uint256()
        self._withdrawalRequestIdsByProvider = mapping(lambda: EnumerableSet());
        self._withdrawalRequests = mapping(lambda: WithdrawalRequest());

        self._setLockDuration(self.DEFAULT_LOCK_DURATION);

    # solhint-enable func-name-mixedcase

    '''
     * @inheritdoc Upgradeable
    '''
    def version(self) -> (int):
        return 4;

    '''
     * @inheritdoc IPendingWithdrawals
    '''
    def lockDuration(self) -> (uint):
        return self._lockDuration.clone();

    '''
     * @dev sets the lock duration
     *
     * notes:
     *
     * - updating it will affect existing locked positions retroactively
     *
     * requirements:
     *
     * - the caller must be the admin of the contract
    '''
    def setLockDuration(self, newLockDuration) -> None:
        self._setLockDuration(newLockDuration);

    '''
     * @inheritdoc IPendingWithdrawals
    '''
    def withdrawalRequestCount(self, provider) -> (int):
        return self._withdrawalRequestIdsByProvider[provider].length();

    '''
     * @inheritdoc IPendingWithdrawals
    '''
    def withdrawalRequestIds(self, provider) -> (list):
        return self._withdrawalRequestIdsByProvider[provider].values();

    '''
     * @inheritdoc IPendingWithdrawals
    '''
    def withdrawalRequest(self, id) -> (WithdrawalRequest):
        return self._withdrawalRequests[id];

    '''
     * @inheritdoc IPendingWithdrawals
    '''
    def initWithdrawal(self,
        provider,
        poolToken,
        poolTokenAmount
    ) -> (uint):
        return self._initWithdrawal(provider, poolToken, poolTokenAmount);

    '''
     * @inheritdoc IPendingWithdrawals
    '''
    def cancelWithdrawal(self, provider, id) -> (uint):
        request = self._withdrawalRequests[id];

        if (request.provider != provider):
            revert("AccessDenied");

        return self._cancelWithdrawal(request, id);

    '''
     * @inheritdoc IPendingWithdrawals
    '''
    def completeWithdrawal(self,
        contextId,
        provider,
        id
    ) -> (CompletedWithdrawal):
        request = self._withdrawalRequests[id];

        if (provider != request.provider):
            revert("AccessDenied");

        currentTime = self._time();
        if (not self._canWithdrawAt(currentTime, request.createdAt)):
            revert("WithdrawalNotAllowed");

        # remove the withdrawal request and its id from the storage
        self._removeWithdrawalRequest(provider, id);

        # approve the caller to transfer the locked pool tokens
        request.poolToken.approve(self.msg_sender, request.poolTokenAmount);

        return CompletedWithdrawal({
            'poolToken': request.poolToken,
            'poolTokenAmount': request.poolTokenAmount,
            'reserveTokenAmount': request.reserveTokenAmount
        });

    '''
     * @inheritdoc IPendingWithdrawals
    '''
    def isReadyForWithdrawal(self, id) -> (bool):
        request = self._withdrawalRequests[id];

        return request.provider != address(0) and self._canWithdrawAt(self._time(), request.createdAt);

    '''
     * @dev sets the lock duration
     *
     * notes:
     *
     * - updating it will affect existing locked positions retroactively
     *
    '''
    def _setLockDuration(self, newLockDuration) -> None:
        prevLockDuration = self._lockDuration;
        if (prevLockDuration == newLockDuration):
            return;

        self._lockDuration = uint32(newLockDuration);

    '''
     * @dev initiates liquidity withdrawal
    '''
    def _initWithdrawal(self,
        provider,
        poolToken,
        poolTokenAmount
    ) -> (uint):
        # record the current withdrawal request alongside previous pending withdrawal requests
        id = self._nextWithdrawalRequestId.clone(); self._nextWithdrawalRequestId += 1;

        # get the pool token value in reserve/pool tokens
        pool = poolToken.reserveToken();
        reserveTokenAmount = self._poolTokenToUnderlying(pool, poolTokenAmount);
        self._withdrawalRequests[id] = WithdrawalRequest({
            'provider': provider,
            'poolToken': poolToken,
            'reserveToken': pool,
            'poolTokenAmount': poolTokenAmount,
            'reserveTokenAmount': reserveTokenAmount,
            'createdAt': self._time()
        });

        if (not self._withdrawalRequestIdsByProvider[provider].add(id)):
            revert("AlreadyExists");

        return id;

    '''
     * @dev returns the pool token value in tokens
    '''
    def _poolTokenToUnderlying(self, pool, poolTokenAmount) -> (uint):
        if (pool is (self._bnt)):
            return self._bntPool.poolTokenToUnderlying(poolTokenAmount);

        return self._network.collectionByPool(pool).poolTokenToUnderlying(pool, poolTokenAmount);

    '''
     * @dev cancels a withdrawal request
    '''
    def _cancelWithdrawal(self, request, id) -> (uint):
        # remove the withdrawal request and its id from the storage
        self._removeWithdrawalRequest(request.provider, id);

        # transfer the locked pool tokens back to the provider
        request.poolToken.transfer(request.provider, request.poolTokenAmount);

        return request.poolTokenAmount;

    '''
     * @dev removes withdrawal request
    '''
    def _removeWithdrawalRequest(self, provider, id) -> None:
        if (not self._withdrawalRequestIdsByProvider[provider].remove(id)):
            revert("DoesNotExist");

        del self._withdrawalRequests[id];

    '''
     * @dev returns whether it's possible to withdraw a request at the provided time
    '''
    def _canWithdrawAt(self, time, createdAt) -> (bool):
        return createdAt + self._lockDuration <= time;



# coding=utf-8
# --------------------------------------------------------------------------------------------------------------------
# Licensed under the MIT LICENSE. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------------------------------
"""Spec and agent-based simulation modules for Bancor v3."""

import warnings

import bancor_research.bancor_simulator.v3.spec

with warnings.catch_warnings():
    warnings.simplefilter("ignore", FutureWarning)
    warnings.simplefilter("ignore", DeprecationWarning)

warnings.filterwarnings("ignore", category=FutureWarning)
warnings.filterwarnings("ignore", category=DeprecationWarning)

__version__ = "3.1.1"
